#include <stdio.h>
#include <i86.h>
#include <dos.h>

/*
  9. 12. 2023: Support for larger than 2 GB disks
*/

typedef unsigned long seekpos_t;
typedef unsigned long lbaindex_t;

unsigned long clock_ticks(void);
#pragma aux clock_ticks = \
  "push ax" \
  "xor ah, ah" \
  "int 0x1A" \
  "pop ax" \
  value [cx dx];

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];

struct drive_struct
{
  unsigned long last_access;
  lbaindex_t block_pos;
  unsigned block_offset;
  unsigned block_size;
  unsigned char drive_num;
};

#define CACHE_SIZE 16

struct drive_struct __far * __based(__segname("_CODE")) drive_struct;
unsigned __based(__segname("_CODE")) drive_count = 0;

struct block_cache
{
  lbaindex_t block;
  unsigned seg;
  unsigned refs;
  unsigned char drive;
};
struct block_cache __based(__segname("_CODE")) block_caches[CACHE_SIZE];

int get_parameters(unsigned, unsigned*);
int detect(unsigned);

void get_dos_dmabuffer(void *ret);
#pragma aux get_dos_dmabuffer = \
  "mov ah, 0x6a" \
  "mov cx, 6" \
  "int 0x21" \
  parm [di];

unsigned current_psp(void);
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

void __far *dos_alloc(unsigned bytes);
#pragma aux dos_alloc = \
  "push cx" \
  "mov cl, 4" \
  "add bx, 0xF" \
  "shr bx, cl" \
  "pop cx" \
  "mov ah, 0x48" \
  "int 0x21" \
  "jnc end" \
  "xor ax, ax" \
  "end:" \
  "xor dx, dx" \
  "xchg ax, dx" \
  parm [bx] value [dx ax];

void dos_free(unsigned seg);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

unsigned serial_receive(unsigned devnum);
#pragma aux serial_receive = \
  "mov ah, 2" \
  "int 0x14" \
  value [ax] parm [dx];

void memset(void __far *dst, unsigned char chr, unsigned count);
#pragma aux memset = \
  "rep stosb" \
  parm [es di] [ax] [cx];

void memcpy(void __far *dst, void __far *src, unsigned count);
#pragma aux memcpy = \
  "push ds" \
  "mov ds, dx" \
  "rep movsb" \
  "pop ds" \
  parm [es di] [dx si] [cx];

unsigned char __based(__segname("_CODE")) error_code = 0;

unsigned kernel_allocate_usercache(unsigned, void*);
#pragma aux kernel_allocate_usercache = \
  "push ds" \
  "mov ax, cs" \
  "mov ds, ax" \
  "mov ax, 0x6d02" \
  "int 0x21" \
  "pop ds" \
  "jnz noerror" \
  "mov ax, -1" \
  "noerror:" \
  parm [bx] [di] modify [ax bx] value [ax];

static int allocate_cache(unsigned char drive)
{
  unsigned n;
  int rc = 0;
  unsigned lr = -1;

  for(n=CACHE_SIZE;n--;)
  {
    if(block_caches[n].drive == 0xFF)
    {
      rc = n;
      break;
    }
    if(block_caches[n].refs < lr)
    {
      lr = block_caches[n].refs;
      rc = n;
    }
  }
  if(block_caches[rc].drive != 0xFF)
    goto allocated;

  n = kernel_allocate_usercache(2048>>4, &block_caches[rc].drive);
  if(n != -1)
  {
    block_caches[rc].seg = n;
    allocated:
    block_caches[rc].drive = drive;
    block_caches[rc].refs = 16;
    return rc;
  }
  else return -1;
}

static int find_cache_block(unsigned char drive, lbaindex_t block)
{
  int n;
  int rc = -1;
  for(n=CACHE_SIZE;n--;)
    if(block_caches[n].drive == drive && block_caches[n].block == block)
    {
      block_caches[n].refs+=2;
      rc = n;
    }
    else
      if(block_caches[n].refs)
        block_caches[n].refs--;
  return rc;
}

static void invalidate_cache(unsigned char drive)
{
  unsigned n;
  for(n=CACHE_SIZE;n--;)
    if(block_caches[n].drive == drive)
      block_caches[n].drive = -1;
}

static void save_to_cache(unsigned char drive, lbaindex_t block, void __far *data)
{
  int c = find_cache_block(drive, block);

  if(c == -1)
  {
    c = allocate_cache(drive);
    if(c == -1) return;
  }
  else return;
  block_caches[c].block = block;
  memcpy(MK_FP(block_caches[c].seg, 0), data, 2048);
}

static void create_lba_packet(lbaindex_t index, void __far *buff,
                              unsigned char *packet, unsigned sectors)
{
  memset(packet, 0, 0x10);
  packet[0] = 0x10;
  packet[2] = sectors;
  *(void __far**)&packet[4] = buff;
  *(lbaindex_t*)&packet[8] = index;
}

unsigned read_lba(unsigned char disk_id, void *packet);
#pragma aux read_lba = \
  "mov ah, 0x42" \
  "int 0x13" \
  "mov al, ah" \
  parm [dl] [si] value [ax];

unsigned seek_lba(unsigned char disk_id, void *packet);
#pragma aux seek_lba = \
  "mov ah, 0x47" \
  "int 0x13"  \
  "mov al, ah" \
  parm [dl] [si] value [ax];

static unsigned __far cd_open(unsigned, unsigned);
#pragma aux cd_open value [ax] parm [bx] [cx] modify exact [ax];
static unsigned __far cd_open(unsigned devnum, unsigned mode)
{
  if(devnum >= drive_count) return -1;
  if(mode) error_code = 5;
  return 0;
}

static void __far * read_blocks(unsigned devnum, lbaindex_t sector, unsigned __far *count)
{
  unsigned char dbuff[6];
  unsigned char lba_packet[16];
  #define dma_buff (*(unsigned char __far**)&dbuff[2])
  #define max_sectors ((*(unsigned *)&dbuff[0])>>11)
  #define sector_size 2048
  int rc;
  unsigned char retries;
  unsigned n;

  get_dos_dmabuffer(dbuff);
  if(*count > max_sectors)
    *count = max_sectors;
  for(retries=3;retries--;)
  {
    create_lba_packet(sector, dma_buff, lba_packet, *count);
    if(!(rc = read_lba(drive_struct[devnum].drive_num, lba_packet)))
      break;
  }
  if(rc)
  {
    error_code = 0x15;
    *count = lba_packet[2];
  }
  for(n=*count;n--;)
    save_to_cache(devnum, sector+n, &dma_buff[sector_size*n]);
  return dma_buff;
}

static void __far *fix_farptr_offset(unsigned char __far *ptr)
{
  unsigned segment = FP_SEG(ptr);
  unsigned offset = FP_OFF(ptr);
  segment += offset >> 4;
  offset &= 0xF;
  return MK_FP(segment, offset);
}

static unsigned __far cd_read(unsigned, unsigned char __far*, unsigned);
#pragma aux cd_read value [ax] parm [bx] [dx ax] [cx] modify exact [ax];
static unsigned __far cd_read(unsigned devnum, unsigned char __far *dst, unsigned count)
{
  unsigned retval = 0;
  lbaindex_t sector = drive_struct[devnum].block_pos;
  unsigned offset = drive_struct[devnum].block_offset;
  unsigned copy_count;
  unsigned sector_count;
  int c;
  unsigned char __far *copy_src;

  if(!count) return 0;

  while(1)
  {
    if((c = find_cache_block(devnum, sector)) != -1)
    {
      sector_count = 1;
      copy_src = MK_FP(block_caches[c].seg, 0);
    }
    else
    {
      sector_count = ((offset + count) >> 11) + ((offset + count) & sector_size - 1);
      copy_src = read_blocks(devnum, sector, &sector_count);
      if(!sector_count) return 0;
    }
    copy_count = sector_size * sector_count - offset;
    if(count < copy_count) copy_count = count;
    dst = fix_farptr_offset(dst);
    memcpy(dst, &copy_src[offset], copy_count);
    dst += copy_count;
    retval += copy_count;
    count -= copy_count;
    if(!count) break;
    sector+=sector_count;
    offset = 0;
  }

  drive_struct[devnum].block_pos += retval >> 11;
  drive_struct[devnum].block_offset += offset + copy_count;
  if(drive_struct[devnum].block_offset > sector_size)
  {
    drive_struct[devnum].block_pos++;
    drive_struct[devnum].block_offset &= sector_size-1;
  }

  drive_struct[devnum].last_access = clock_ticks();

  return retval;
}

void debug(unsigned char chr);
#pragma aux debug = \
  "mov ah, 0x0E" \
  "mov bx, 0x0007" \
  "int 0x10" \
  parm [al] modify [bx];

void int3(void);
#pragma aux int3 = \
  "int 3";


static unsigned long __far cd_seek(unsigned devnum, unsigned long count, unsigned origin);
#pragma aux cd_seek parm [bx] [dx ax] [cx] value [dx ax] modify exact [dx ax];
static unsigned long __far cd_seek(unsigned devnum, unsigned long count, unsigned origin)
{
  unsigned offset = count & (sector_size - 1);
  switch(origin)
  {
    case 0:
      drive_struct[devnum].block_pos = count >> 11;
      drive_struct[devnum].block_offset = offset;
      return count;
    case 1:
      drive_struct[devnum].block_pos += count >> 11;
      if((drive_struct[devnum].block_offset += offset) > sector_size)
      {
        drive_struct[devnum].block_pos++;      
        drive_struct[devnum].block_offset &= sector_size - 1;
      }
      return (drive_struct[devnum].block_pos << 11)
           + drive_struct[devnum].block_offset;
    case 2:
      drive_struct[devnum].block_offset = -1;
      drive_struct[devnum].block_pos = -1;
      return -1;
  }
}

unsigned __far get_error(unsigned devnum);
#pragma aux get_error parm [bx] value [ax] modify exact [ax];
unsigned __far get_error(unsigned devnum)
{
  unsigned retval = error_code;
  error_code = 0;
  return retval;
}

int __far driver_command(unsigned, unsigned short __far*);
#pragma aux driver_command parm [bx] [dx ax] value [ax] modify exact [ax];
int __far driver_command(unsigned devnum, unsigned short __far *cmd)
{
  if(clock_ticks() - drive_struct[devnum].last_access > (5*14))
  {
    invalidate_cache(devnum);
    return 1;
  }
  return 0;
}

unsigned long __far driver_ioctl(unsigned, unsigned, unsigned char __far*, unsigned);
#pragma aux driver_ioctl parm [bx] [cx] [dx ax] [si] value [dx ax] modify exact [dx ax];

#define CHAR_DEVICE 0x80L
#define NO_EOF_ON_INPUT 0x40L
#define BINARY_MODE 0x20L

unsigned long __far driver_ioctl(unsigned devnum, unsigned count,
                                 unsigned char __far *buff, unsigned function)
{
  unsigned return_word;
//  unsigned char eof;
//  eof = (serial_cb[devnum].write_offset == serial_cb[devnum].read_offset);
  switch(function & 0xFF)
  {
    case 0: case 1:
      return_word = CHAR_DEVICE | BINARY_MODE;
//      if(!eof)
//        return_word |= NO_EOF_ON_INPUT;
      return ((long)return_word << 16) | function;
    case 6:
//      if(eof)
//        return ((long)FP_SEG(buff) << 16) | (function & 0xFF00);
//      else
        return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
    case 7:
        return ((long)FP_SEG(buff) << 16) | (function & 0xFF00);
  }
}

#define FIRST_CD_DEV 0x80
int cd_initialize(void)
{
  unsigned char devnum;
  int rc;
  unsigned char devnums[0x100 - FIRST_CD_DEV];
  unsigned n;

//  for(devnum=0x80;devnum;devnum++)
  for(devnum=0xE0;devnum<0xF0;devnum++)
  {
    rc = detect(devnum);
    if(!rc)
    {
      devnums[drive_count] = devnum;
      drive_count++;
    }
  }
  if(drive_count)
  {
    drive_struct = dos_alloc(drive_count * sizeof(struct drive_struct));
    if(!drive_struct)
    {
      puts("Error: Out of memory while allocating drive structures!");
      return -1;
    }
    else
    {
      for(n=drive_count;n--;)
      {
        drive_struct[n].drive_num = devnums[n];
        drive_struct[n].block_pos = 0;
        drive_struct[n].block_offset = 0;
        drive_struct[n].last_access = clock_ticks()-(5*14);
      }
      printf("Allocated drive structures for %u CDROM drives.\n", drive_count);
    }
  }
  else
  {
    puts("No CDROM drives detected.");
    return -1;
  }

  for(n=CACHE_SIZE;n--;)
    block_caches[n].drive = -1;

  return 0;
}


int install_driver(char __far *, void *);
#pragma aux install_driver = \
  "mov ah, 0x6A" \
  "xor cx, cx" \
  "push ds" \
  "push cs" \
  "pop ds" \
  "int 0x21" \
  "pop ds" \
  "jnc noerror" \
  "mov ax, -1" \
  "jmp end" \
  "noerror:" \
  "xor ax, ax" \
  "end:" \
  parm [dx bx] [di] value [ax] modify [cx];

  void __far * __based(__segname("_CODE")) function_table[] =
  {
    cd_read,
    0, // write
    cd_seek, // seek
    cd_open,
    0, // close
    get_error, // get_error_code
    driver_command, // send_command
    driver_ioctl // ioctl
  };

void set_memory_allocation_strategy(unsigned strategy);
#pragma aux set_memory_allocation_strategy = \
  "mov ax, 0x5801" \
  "int 0x21" \
  parm [bx];

int main()
{
  unsigned env_seg;
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  int rc;
  unsigned buf[3];
  char drivername[] = "CD";
  unsigned n;

  puts("Installing CDROM driver v1.1.");

  get_dos_dmabuffer(buf);
  if(buf[0] < 2048)
  {
    puts("Seems that your computer doesn't have enough memory for 2048-byte \
sectors. :(");
    return -1;
  }
  printf("Using kernel DMA buffer in address %.4X:%.4X\n", buf[2], buf[1]);

  set_memory_allocation_strategy(2);
  rc = cd_initialize();
  set_memory_allocation_strategy(0);
  if(rc) return 1;

  if(!install_driver(drivername, function_table))
  {
    puts("Done, going resident...");

    env_seg = *(unsigned __far*)&psp_ptr[0x2C];
    dos_free(env_seg);

    for(n=3;n--;)
      dos_close(n);

    _dos_keep(0, 0x10 + ((FP_OFF(main) + 0xF) >> 4));
  }
  puts("Error");
  return 1;
}

#include "drdetect.c"
