#include <stdio.h>
#include <stdlib.h>
#include <i86.h>

int extensions_present(unsigned drive);
#pragma aux extensions_present = \
  "mov ah, 0x41" \
  "mov bx, 0x55AA" \
  "int 0x13" \
  "jc no_extensions" \
  "cmp bx, 0xAA55" \
  "jne no_extensions" \
  "cmp ah, 0x21" \
  "jl no_extensions" \
  "mov ax, cx" \
  "jmp end" \
  "no_extensions:" \
  "xor ax, ax" \
  "end:" \
  parm [dx] value [ax] modify [ax bx cx];

int bios_get_parameters(unsigned drive, void *buff);
#pragma aux bios_get_parameters = \
  "mov ah, 0x48" \
  "int 0x13" \
  "lahf" \
  "and ah, 0x01" \
  "mov al, ah" \
  parm [dx] [si] value [ax];

int get_parameters(unsigned drive, unsigned *ret_flags)
{
  unsigned char buff[30] = { 30, 00 };
  int rc = bios_get_parameters(drive, buff);
  unsigned *flags = (unsigned*)&buff[2];
  #define conf_params (*(unsigned char __far**)&buff[26])
  #define hardware_flags *(unsigned __far*)&conf_params[10]
  #define io_port *(unsigned __far*)&conf_params[0]
  #define control_port *(unsigned __far*)&conf_params[2]
  #define irq (conf_params[6] & 0xF)

  if(rc)
  {
    printf("Error while getting parameters of device 0x%X.\n", drive);
    return rc;
  }
  else
  {
    *ret_flags = *flags;
    printf("Parameters of device 0x%X:\n", drive);
    if(*flags & 0x04)
      printf("Device is removable.\n");
    if(*flags & 0x08)
      printf("Device has change line support.\n");
    if(*flags & 0x10)
      printf("Device is lockable.\n");
    if(*flags & 0x20)
      printf("No media present.\n");
    printf("Pointer to configuration parameters: %.4X:%.4X\n", FP_SEG(conf_params), FP_OFF(conf_params));
    if(conf_params != (void __far*)-1)
    {
      printf("Hardware flags: %.4X\n", hardware_flags);
      if(hardware_flags & 0x40)
        printf("Device is an ATAPI device\n");
      printf("I/O port: %X, Control port: %X, IRQ: %X\n",
               io_port, control_port, irq);
    }
  }
  return 0;
}

int detect(unsigned drive)
{
  unsigned flags;
  int rc = extensions_present(drive);
  if(rc)
  {
    printf("Drive 0x%X supports BIOS int 13h extensions (0x%X).\n", drive, rc);
    if(!(rc & 0x01))
    {
      puts("No packet structure supported - cannot read this drive!");
      return -1;
    }
    if(get_parameters(drive, &flags))
    {
      printf("Error while trying to get parameters for drive 0x%X\n", drive);
      return -1;
    }
    if(flags & 0x04)
    {
      printf("Drive 0x%X is removable - using as CD%u device...\n", drive, drive_count);
      return 0;
    }
  }
  return -1;
}
