#define SECTOR_SIZE 2048

unsigned char call_critical_error(unsigned drive, unsigned error_code);
#pragma aux call_critical_error = \
  "mov ah, 0x7E" \
  "call dword ptr cs:kernel_criterr_hdl" \
  parm [ax] [di] value [al];

inline unsigned char critical_error(unsigned char drive, unsigned char error)
{
  error = bios2doserror(error);
  return call_critical_error(drive, error);
}

unsigned char bios_error_code(void);
#pragma aux bios_error_code = \
  "mov ah, 1" \
  "int 0x13" \
  value [al];

void int_23h(void);
#pragma aux int_23h = \
  "int 0x23";

static unsigned read_drive(struct fs_status __far *fs_status, void __far *buf, unsigned count)
{
  unsigned rc;
  read:rc = dos_read(fs_status->handle, buf, count);
  if(error_code && !fs_status->image)
    switch(critical_error(fs_status->drive, bios_error_code()))
    {
      case 0:
        error_code = 0;
        break;
      case 1:
        error_code = 0;
        dos_seek(fs_status->handle, -rc, 1); 
        goto read;
      case 2: int_23h();
//      case 3:
    }
  return rc;
}

int __far fs_changed(unsigned char drive);
#pragma aux fs_changed parm [bx] value [ax] modify exact [ax];

int set_partition_size_pragma(void __far*);
#pragma aux set_partition_size_pragma = \
  "push ds" \
  "push cx" \
  "mov ax, es" \
  "mov ds, ax" \
  "mov cx, 7" \
  "mov ah, 0x6a" \
  "int 0x21" \
  "pop cx" \
  "pop ds" \
  parm [es di] value [ax];

static int set_partition_size(struct fs_status __far *fs_status)
{
  unsigned char strct[12];

  strct[0] = fs_status->drive;
  *(unsigned __based(__segname("_STACK"))*)&strct[2] = 2048;
  *(unsigned long __based(__segname("_STACK"))*)&strct[4] = fs_status->block_count;

  return set_partition_size_pragma(strct);
}


static int initialize_disk(struct fs_status __far *fs_status)
{
  dos_seek(fs_status->handle, 0x8000, 0);
  if(read_drive(fs_status, work_buff, SECTOR_SIZE) != SECTOR_SIZE)
  {
    error_code = 0x0D;
    return -1;
  }
  if(error_code)
    return -1;

  fs_status->block_count = *(unsigned long __far*)&work_buff[80];
  fs_status->rootdir_start_block = *(unsigned long __far*)&work_buff[156+2];
  fs_status->rootdir_length = *(unsigned long __far*)&work_buff[156+10];

  set_partition_size(fs_status);

  return 0;
}

static int mount_fs(struct fs_status __far *fs_status)
{
  if(initialize_disk(fs_status) == -1)
    return -1;
  fs_status->workdir_start_block = 0;

  return 0;
}

static int save_workdir_start_block(struct fs_status __far*);

static int seek2direntry(struct fs_status __far *fs_status, dirindex_t index)
{
  unsigned char record_size;
  unsigned long o = 0;
  unsigned rc;
  unsigned read_count;
  unsigned n;

  if(!fs_status->workdir_start_block)
    save_workdir_start_block(fs_status);

  do
  {
    if(o >= fs_status->workdir_length) return -2;
    read_count = fs_status->workdir_length - o;
    if(read_count > 256) read_count = 256;
    seek_sector(fs_status->handle, fs_status->workdir_start_block);
    dos_seek(fs_status->handle, o, 1);
    rc = read_drive(fs_status, work_buff, read_count);
    if(error_code) return -3;
    if(!rc) return -2;
    record_size = work_buff[0];
    if(!record_size)
    {
      n = SECTOR_SIZE - (o & (SECTOR_SIZE-1));
      o += n;
      continue;
    }
    o += record_size;
  }
  while(index--);

  return 0;
}

static int save_workdir_start_block(struct fs_status __far *fs_status)
{
  unsigned n;
  int rc;

  fs_status->workdir_start_block = fs_status->rootdir_start_block;
  fs_status->workdir_length = fs_status->rootdir_length;

  for(n=0;fs_status->working_directory[n] != -1;n++)
  {
    if(rc = seek2direntry(fs_status, fs_status->working_directory[n]))
    {
      fs_status->workdir_start_block = 0;
      fs_status->workdir_length = 0;
      return rc;
    }
    fs_status->workdir_start_block = *(unsigned long __far*)&work_buff[2];
    fs_status->workdir_length = *(unsigned long __far*)&work_buff[10];
  }
  return 0;
}


static int get_direntry(struct fs_status __far *fs_status,
                        dirindex_t index,
                        unsigned char __far *buff)
{
  unsigned char name_len;
  int rc;
  unsigned char __far *name;

  if(fs_changed(fs_status->drive) == -1) return -3;

  rc = seek2direntry(fs_status, index);
  if(rc)
  {
    return rc;
  }

  name_len = work_buff[32];
  if(name_len > 12) name_len = 12;
  name = &work_buff[33];

  if(*name > 1)
    while(name_len-- && *name != ';')
      *buff++ = *name++;
  else
  {
    *buff++ = '.';
    if(*name == 1)
      *buff++ = '.';
  }

  *buff = 0;

  return 0;
}

static int directory_entry(unsigned char drive, dirindex_t index,
                           unsigned char __far *buff, unsigned char count)
{
  struct fs_status __far *fs_status = fs_struct(drive);

  return get_direntry(fs_status, index, buff);
}

static int fileinfo(unsigned char drive, dirindex_t index, struct file_info *ret)
{
  struct fs_status __far *fs_status = fs_struct(drive);
  unsigned flags;

  seek2direntry(fs_status, index);

  ret->size = *(unsigned long __far*)&work_buff[10];
  flags = work_buff[25];
  ret->attr = 0x01;
  if(flags & 0x01) ret->attr |= 0x02;
  if(flags & 0x02) ret->attr |= 0x10;
  ret->date.year = work_buff[18]+1900;
  ret->date.month = work_buff[19];
  ret->date.day = work_buff[20];
  ret->time.hour = work_buff[21];
  ret->time.minute = work_buff[22];
  ret->time.second = ret->time.ssecond = 0;


  return 0;
}

static void free_space(unsigned char drive, struct diskfree *retval)
{
  struct fs_status __far *fs_status = fs_struct(drive);

//  retval->clusters = fs_status->block_count;
  retval->clusters = fs_status->block_count >> 7;
  retval->bytes_per_sector = 2048;
  retval->unused_clusters = 0;
//  retval->sectors_per_cluster = 1;
  retval->sectors_per_cluster = 128;
}

static void chdir(unsigned char drive, dirindex_t __far* index)
{
  struct fs_status __far *fs_status = fs_struct(drive);
  unsigned n;

  fs_status->workdir_start_block = 0;
  fs_status->workdir_length = 0;
  for(n=0;;n++)
  {
    if((fs_status->working_directory[n] = index[n]) == -1)
      return;
  }
}

static int getcwd(unsigned char drive, dirindex_t __far *index)
{
  struct fs_status __far *fs_status = fs_struct(drive);
  int rc = 0;

  while(1)
  {
    index[rc] = fs_status->working_directory[rc];
    if(index[rc++] == -1)
      return rc;
  }
}

static unsigned long gethandlestaticinfo(unsigned char drive, dirindex_t index)
{
  struct fs_status __far *fs_status = fs_struct(drive);
  int rc;

  rc = seek2direntry(fs_status, index);
  if(rc)
  {
    error_code = 0x0E;
    return 0;
  }

  return *(unsigned long __far*)&work_buff[2];
}

static unsigned read_file(unsigned long offset, dirindex_t __far *file_index, 
                   unsigned count, void __far *buff, unsigned handle_seg)
{
  unsigned char drive = *file_index++;
  struct fs_status __far *fs_status = fs_struct(drive);
  unsigned long __far *handle_ptr = MK_FP(handle_seg, 0);
  unsigned file_index_len;
  dirindex_t index;
  unsigned long start_block;
  unsigned long file_len;
  dirindex_t old_cwd[MAX_DIR_DEPTH+1];

  if(handle_seg)
  {
    start_block = handle_ptr[1];
    if(!start_block)
      goto getstartblock;
    file_len = handle_ptr[2];
  }
  else
  {
    getstartblock:
    for(file_index_len=0;file_index[file_index_len] != -1;file_index_len++);
    file_index_len--;
    index = file_index[file_index_len];
    file_index[file_index_len] = -1;

    getcwd(drive, old_cwd);
    chdir(drive, file_index);

    seek2direntry(fs_status, index);
    start_block = *(unsigned long __far*)&work_buff[2];
    file_len = *(unsigned long __far*)&work_buff[10];

    file_index[file_index_len] = index;
    chdir(drive, old_cwd);
  }

  if(offset > file_len) return 0;
  if(file_len < offset+count)
    count = file_len-offset;

  seek_sector(fs_status->handle, start_block + (offset >> 11));
  dos_seek(fs_status->handle, offset & (2048-1), 1);
  return read_drive(fs_status, buff, count);
}
