#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

const char comname[] = "COM";
const char lptname[] = "LPT";
const char conname[] = "CON";

const char errorstr[] = "Error!";
const char nosuchdevice[] = "No such device!";

unsigned short baud_divisors[] = 
{
      // baud rates:  
2304, // 0 == 50 baud
1040, // 1 == 110 baud
768,  // 2 == 150 baud
384,  // 3 == 300 baud
192,  // 4 == 600 baud
96,   // 5 == 1200 baud
48,   // 6 == 2400 baud
24,   // 7 == 4800 baud
12,   // 8 == 9600 baud
6,    // 9 == 19200 baud
3    // 10 == 38400 baud
};

unsigned short bauds[] =
{
  50,
  110,
  150,
  300,
  600,
  1200,
  2400,
  4800,
  9600,
  19200,
  38400
};


#define XON_XOFF 0x01
#define HARDWARE 0x02

int command_driver(char __far *devname, void __far *cmd);
#pragma aux command_driver = \
  "push ds" \
  "push es" \
  "pop ds" \
  "mov ah, 0x6B" \
  "int 0x21" \
  "pop ds" \
  parm [dx bx] [es di] value [ax];

int set_serial_parms(int argc, char *argv[], char *comport_name)
{
  unsigned baud;
  char *flowctrls[] =
  {
    "NONE", "SOFT", "HARD"
  };
  unsigned short driver_cmd[6];
  #define CMD_OPCODE driver_cmd[0]
  #define PARM_BAUD driver_cmd[1]
  #define PARM_PARITY driver_cmd[2]
  #define PARM_WORDSIZE driver_cmd[3]
  #define PARM_STOPBITS driver_cmd[4]
  #define PARM_FLOWCTRL driver_cmd[5]

  if(argc >= 4)
  {
    CMD_OPCODE = 1; // set parameters
    baud = atoi(argv[0]);
    PARM_PARITY = argv[1][0];
#define PARM_PARITY_ODD (1 << 3)
#define PARM_PARITY_EVEN (3 << 3)
#define PARM_PARITY_MARK (5 << 3)
#define PARM_PARITY_SPACE (7 << 3)
    switch(PARM_PARITY)
    {
      case 'o': PARM_PARITY = PARM_PARITY_ODD; break;
      case 'e': PARM_PARITY = PARM_PARITY_EVEN; break;
      case 'm': PARM_PARITY = PARM_PARITY_MARK; break;
      case 's': PARM_PARITY = PARM_PARITY_SPACE; break;
      case 'n': PARM_PARITY = 0; break;
      default:
        puts("Invalid parity");
        return 1;
    }
    PARM_WORDSIZE = atoi(argv[2]);
    if(PARM_WORDSIZE < 5 || PARM_WORDSIZE > 8)
    {
      puts("Invalid word size");
      return 1;
    }
    PARM_STOPBITS = atoi(argv[3]);
    if(PARM_STOPBITS > 2)
    {
      puts("Too many stop bits");
      return 1;
    }
    if(argc >= 5)
    {
      if(!strcmp(argv[4], flowctrls[XON_XOFF]))
        PARM_FLOWCTRL = XON_XOFF;
      else if(!strcmp(argv[4], flowctrls[HARDWARE]))
        PARM_FLOWCTRL = HARDWARE;
      else if(!strcmp(argv[4], flowctrls[0]))
        goto noflowctrl;
      else
      {
        puts("Invalid flow control");
        return 1;
      }
    }
    else noflowctrl:PARM_FLOWCTRL = 0;

    for(PARM_BAUD=sizeof(bauds)/sizeof(unsigned short);PARM_BAUD--;)
      if(bauds[PARM_BAUD] == baud) break;
    if(PARM_BAUD == -1)
    {
      puts("Invalid baud");
      return 1;
    }
//    PARM_BAUD = baud_divisors[PARM_BAUD];
  }
  else CMD_OPCODE = 0;

  switch(command_driver(comport_name, driver_cmd))
  {
    case 0: break;
    case 2:
      puts(nosuchdevice);
    default:
      puts(errorstr);
      return 1;
  }

  baud = bauds[PARM_BAUD];

  if(PARM_PARITY == PARM_PARITY_ODD) PARM_PARITY = 'o';
  else if(PARM_PARITY == PARM_PARITY_EVEN) PARM_PARITY = 'e';
  else if(PARM_PARITY == PARM_PARITY_MARK) PARM_PARITY = 'm';
  else if(PARM_PARITY == PARM_PARITY_SPACE) PARM_PARITY = 's';
  else PARM_PARITY = 'n';

  printf("%s: BAUDS=%u PARITY=%c WORDSIZE=%u STOPBITS=%u FLOWCTRL=%s\n",
    comport_name, baud, PARM_PARITY, PARM_WORDSIZE, PARM_STOPBITS, flowctrls[PARM_FLOWCTRL]);

  return 0;
} 

int set_parallel_parms(int argc, char *argv[], char *lptport_name)
{
  unsigned char driver_cmd[2];
  #define LPT_OPCODE driver_cmd[0]
  #define LPT_LINELEN driver_cmd[1]

  if(argc)
  {
    LPT_OPCODE = 1;
    LPT_LINELEN = atoi(argv[0]);
  }
  else
    LPT_OPCODE = 0;

  switch(command_driver(lptport_name, driver_cmd))
  {
    case 0: break;
    case 2: puts(nosuchdevice);
    default: puts(errorstr);
    return 1;
  }

  printf("LPT1: Line length %u\n", LPT_LINELEN);

  return 0;
}

int set_console_parms(int argc, char *argv[])
{
  unsigned char charh;
  unsigned char cols;
  unsigned char mode;

  if(!argc)
  {
    puts(errorstr);
    return 1;
  }

  cols = atoi(argv[0]);
  __asm
  {
    mov ah, 0x0F
    int 0x10

    xor ah, ah
    cmp cols, 40
    jne hires_txt
    and al, not 0x02
    jmp set_mode
    hires_txt:
    or al, 0x02
    set_mode:
    int 0x10
  }

  if(argc == 2)
  {
    charh = atoi(argv[1]);
    switch(charh)
    {
      case 14:
        mode = 0x11;
        break;
      case 8:
        mode = 0x12;
        break;
      case 16: default:
        return 0;
    }
    __asm
    {
      mov ah, 0x11
      mov al, mode
      xor bl, bl
      int 0x10
    }
  }

  return 0;
}

int istrncmp(char *str1, char *str2, unsigned count)
{
  while(*str1 && *str2)
  {
    if(tolower(*str1) != tolower(*str2))
      break;
    if(!--count)
      return 0;
    str1++;
    str2++;
  }
  return *str1 - *str2;
}

int main(int argc, char *argv[])
{
  if(argc > 1)
  {
    if(!istrncmp(argv[1], (char*)comname, 3))
      return set_serial_parms(argc-2, &argv[2], argv[1]);
    if(!istrncmp(argv[1], (char*)lptname, 3))
      return set_parallel_parms(argc-2, &argv[2], argv[1]);
    if(!istrncmp(argv[1], (char*)conname, 3))
      return set_console_parms(argc-2, &argv[2]);
  }

  puts("Usage:");
  puts("MODE COMn [BAUD] [PARITY(neosm)] [WORDSIZE(5-8)]\
 [STOPBITS(0-2)] [FLOWCTRL(NONE,SOFT,HARD)]");
  puts("MODE LPTn [LINELEN(0 == automatic)]");
  puts("MODE CON [COLS(40, 80)] [CHAR HEIGHT(8, 14, 16)]");
  puts("Examples:");
  puts(">MODE COM1 4800 e 8 1 SOFT");

  return 1;
}
