IFDEF __386__
.386
ELSE
.8086
ENDIF

.MODEL small

public putstr_
public getstr_
public start_program_
public write_con_
public getch_
public putch_
public kbhit_
public load_shell_
public call_idle_handler_
;public abs_
public int15h_handler_
public _bios_int15_handler
public benchmark_
extern print_regs_:BYTE
extern print_sp_:BYTE
extern run_program_:BYTE
extern end_shell:BYTE
extern _code_end:BYTE
extern device_wait_:BYTE
extern get_vector_:BYTE
extern set_vector_:BYTE
extern stack_panic:BYTE
_internal_flags EQU _code_end+57

_TEXT               SEGMENT BYTE PUBLIC USE16 'CODE'
                ASSUME CS:_TEXT                               

kbhit_ proc near
  test byte ptr getch_specialkey, 0xFF
  jz kbhit_getkey
  mov al, 0x01
  ret

  kbhit_getkey:
  mov ax, 0x11FF
  int 0x16
  pushf
  cmp al, 0xFF
  je kbhit_pcxt
  popf
  jnz kbhit_end
  xor ax, ax
  ret

  kbhit_pcxt:
  add sp, 2
  mov ah, 0x01
  int 0x16
  jnz kbhit_end
  xor ax, ax
  kbhit_end:
  ret
kbhit_ endp

; es:si == string to write
putstr_ proc near
  push ax
  push bx
  push si
  push bp ; save BP for bug compatibility with some BIOS implementations

  mov ah, 0x0F
  int 0x10
  mov bl, 0x07

  putstr_loop:
  mov al, es:[si]
  test al, al
  jz putstr_end
  mov ah, 0x0E
  int 0x10
  inc si
  jnz putstr_loop
  mov si, es
  add si, 0x1000
  mov es, si
  xor si, si
  jmp putstr_loop

  putstr_end:

  pop bp
  pop si
  pop bx
  pop ax
  ret
putstr_ endp

; es:si = string, cx = length
write_con_ proc near
  push si
  push bx
  push ax
  push bp
  push cx

  mov ah, 0x0F
  int 0x10
  mov bl, 0x07

  write_con_loop:
  mov ah, 0x0E
  mov al, es:[si]
  int 0x10
  inc si
  jnz write_con_loop_end
  mov si, es
  add si, 0x1000
  mov es, si
  xor si, si
  write_con_loop_end:
  loop write_con_loop

  pop cx
  pop bp
  pop ax
  pop bx
  pop si
  ret
write_con_ endp

call_idle_handler_ proc near
  push cx

;  xor cl, cl  
;  cli
;  xchg cl, byte ptr [_code_end+27] ; unset IN_DOS flag
;  or _internal_flags, 0x02 ; set IN_KERNEL flag
;  int 0x28
;  and _internal_flags, NOT 0x02 ; unset IN_KERNEL flag
;  mov byte ptr [_code_end+27], cl ; restore IN_DOS flag
  xor cx, cx
  cli
  xchg cl, byte ptr [_code_end+27] ; unset IN_DOS flag
  mov ch, _internal_flags  ; save internal flags
  or _internal_flags, 0x02 ; set IN_KERNEL flag
  int 0x28
  mov _internal_flags, ch
  mov byte ptr [_code_end+27], cl

  sti
  pop cx
  ret
call_idle_handler_ endp

getch_specialkey: db 0

getch_ proc near 
  test getch_specialkey, 0xFF
  jz getch_kbhit

  xor al, al
  xchg al, getch_specialkey
  ret

  getch_kbhit:
  call kbhit_
  jnz getch_read_key

  test byte ptr [_code_end+40], 0x08
  jnz getch_in_criterr

  call call_idle_handler_

  hlt  
  getch_in_criterr:

  test byte ptr [_code_end+57], 0x01 ; break key pressed?
  jnz getch_end

  jmp getch_

  getch_read_key:
  mov ax, 0x10FF
  int 0x16
  cmp al, 0xFF
  je getch_pcxt
  jmp getch_testspecial

  getch_pcxt:
  mov ah, 0x00
  int 0x16
  getch_testspecial:
  cmp al, 0xE0
  je getch_special
  test al, al
  jnz getch_end

  getch_special:
  xor al, al
  mov getch_specialkey, ah

  getch_end:
  ret
getch_ endp

putch_ proc near
  push bp
  push bx
  mov ah, 0x0F
  push ax
  int 0x10 ; get active video page
  mov bl, 0x07
  pop ax
  dec ah
  int 0x10
  pop bx
  pop bp
  ret
putch_ endp

; es:di = pointer to buffer
; cx = max length
; returns length of the string to ax
getstr_ proc near
  push di
  push bp
  push si
  push bx
  push cx
  mov si, sp

  mov ah, 0x0F
  int 0x10  ; get active video page
  mov bl, 0x07

  xor cx, cx

  getstr_loop:
  call getch_

  mov ah, 0x0E

  test byte ptr _code_end+57, 0x01 ; break key pressed?
  jnz getstr_end

  cmp al, 8
  je getstr_backspace ;; backspace key

  cmp al, 0x0D
  je getstr_end

  cmp cx, [si]
  je getstr_loop

  int 0x10

  mov es:[di], al
  inc cx
  inc di

  jnz getstr_loop
  mov di, es
  add di, 0x1000
  mov es, di
  xor di, di

  jmp getstr_loop

  getstr_end:
  int 0x10
  mov ax, 0x0E0A
  int 0x10

  mov ax, cx

  pop cx
  pop bx
  pop si
  pop bp
  pop di
  ret

  getstr_backspace:
  ; if cx == 0, do nothing
  jcxz getstr_loop

  ; if cursor is at column 0, jump to previous line
  push cx
  push dx

  mov ah, 0x03
  int 0x10
  test dl, dl
  jnz no_prev_line

  ; how many columns?
  mov ah, 0x0F
  int 0x10
  mov dl, ah
  dec dh

  no_prev_line:
  mov ah, 0x02
  dec dl
  int 0x10

  pop dx
  pop cx

  ; echo whitespace at cursor
  mov ax, 0x0920
  int 0x10

  dec di
  dec cx

  mov byte ptr es:[di], 0x00
  
  jmp getstr_loop
getstr_ endp

; bx = initial cs
; cx = initial ip
; dx = initial ss
; di = initial sp
start_program_ proc far
  ;call near ptr print_regs_
  push bp
  push bx
  push cx
  push dx
  push di
  push si

  ; ds and es = psp
  mov ax, word ptr [_code_end+4]
  mov es, ax
  mov ds, ax

  mov ax, 0xFFFF

  ; switch to program stack
  mov ss, dx
  mov sp, di

  ; if debugging, unset trap flag
IFDEF debug
  pushf
  and word ptr ss:[di-2], NOT 0x0100
ENDIF

  push bx
  push cx

  mov byte ptr cs:[_code_end+27], 0 ; unset IN_DOS flag

IFNDEF debug
  retf
ELSE
  ; if debugging, use iret
  iret
ENDIF

;  mov cs:exe_offset, cx
;  mov cs:exe_segment, bx
;  db 0xEA ; FAR JMP (self-modifying code to not modify the program stack top)
;  exe_offset: dw 0
;  exe_segment: dw 0

start_program_ endp

load_shell_ proc near
;  call near ptr print_sp_
;  call near ptr print_regs_
  push bx
  push cx
  push dx
  push si
  push di
  push bp

  mov word ptr [_code_end+36], sp

  xor cx, cx

  call near ptr run_program_

  jmp near ptr end_shell

load_shell_ endp

;abs_ proc near
;  test ax, 0x8000
;  jnz abs_end
;  not ax
;  inc ax
;  abs_end:
;ret
;abs_ endp

int15h_handler_ proc far
  pushf

  cmp ah, 0x91
  jne int15_handler_1
  cmp byte ptr cs:[_code_end+30], al
  jne int15h_chain_intr
  inc byte ptr cs:[_code_end+31]
  mov byte ptr cs:[_code_end+30], 0xFF
  jmp int15h_chain_intr

  cmp sp, 0x200
  jc int15h_chain_intr

  int15_handler_1:
  cmp ah, 0x90
  jne int15h_chain_intr
  cmp al, 0xC0
  jnc int15h_handler_2
  cmp al, 0x02
  jnc int15h_chain_intr
  int15h_handler_2:
  push ax

  mov word ptr cs:[_code_end+58], sp
  mov word ptr cs:[_code_end+60], ss
  push cs
  pop ss
  mov sp, 0x200
  push ds

  push cs
  pop ds

  xor ah, ah
  mov word ptr [_code_end+30], ax

  push [_code_end+4] ; save CURRENT_PSP
  test word ptr [_code_end+38], 0xFFFF ; has a driver called this int 15h?
  jz int15h_1

  mov ax, word ptr [_code_end+38]
  mov word ptr [_code_end+4], ax ; set CURRENT_PSP to DRVCALL_PSP
  mov ax, word ptr [_code_end+30]

  int15h_1:
  sti
IFDEF __386__
  push fs
  push gs
ENDIF

  call near ptr device_wait_

IFDEF __386__
  pop gs
  pop fs
ENDIF

  pop [_code_end+4] ; restore CURRENT_PSP

  pop ds
  mov ss, word ptr cs:[_code_end+60]
  mov sp, word ptr cs:[_code_end+58]

  int15h_end:
  pop ax

  test cs:[_code_end+31], 0xFF
  jne int15h_chain_intr
  stc
  add sp, 2
  retf 2

  int15h_chain_intr:
  popf

  db 0xEA
  _bios_int15_handler: dd 0
int15h_handler_ endp
;

; parameters: di == pointer to a temporary 16-bit variable
; ax == result
; smaller number == slower machine
benchmark_ proc near
push bx
push cx
push si

mov bx, 0x1C
call near ptr get_vector_
mov word ptr benchmark_timerrtn_old, si
mov word ptr benchmark_timerrtn_old+2, es

mov si, offset benchmark_timerrtn
mov ax, cs
mov es, ax
call near ptr set_vector_

xor ax, ax
mov benchmark_temp, di
hlt
mov word ptr [di], ax
benchmark_loop_2:
mov cx, 0xFFF
benchmark_loop:
test word ptr [di], 0xFFFF
jnz benchmark_loop_end
loop benchmark_loop
inc ax
jnz benchmark_loop_2
benchmark_loop_end:

mov si, word ptr benchmark_timerrtn_old
mov es, word ptr benchmark_timerrtn_old+2
mov bx, 0x1C
call near ptr set_vector_

pop si
pop cx
pop bx
ret

benchmark_timerrtn:
; inc word ptr cs:[di]
db 0x2E
db 0xFF
db 0x06
benchmark_temp: dw 0

db 0xEA
benchmark_timerrtn_old:
dd 0
benchmark_ endp
_TEXT ENDS

END
