#define DEFAULT_REFS 1024
#define ADDREFS 32

/*extern unsigned cache_seg;
extern unsigned rawcache_seg;
extern unsigned rawcache_size;
extern unsigned cache_size;*/

extern char benchmark_res[];

void init_cache(void)
{
  unsigned alloc_size;
  unsigned bm_res = benchmark(&CACHE_SIZE);

  putstr(benchmark_res);
  putstr(hex2str(bm_res >> 8));
  putstr(hex2str(bm_res));

  // 25 MHz 486 == 0x21
  if(bm_res < 4) CACHE_SIZE = 16;
  else if(bm_res < 0x10) CACHE_SIZE = 64;
  else if(bm_res < 0x30) CACHE_SIZE = 128;
  else CACHE_SIZE = 256;


  alloc_size = ((CACHE_SIZE*sizeof(struct cache_entry))+0xF)>>4;

  if(free_paras(0, 0, alloc_size) < alloc_size)
  {
    return;
  }

  cache_seg = rawcache_buff + rawcache_size;
  PARAGRAPH_COUNT(initial_mcb) += alloc_size;

  nmemset(DISK_CACHE, 0, CACHE_SIZE*sizeof(struct cache_entry));

  allocate_more_caches();
}

void disable_caches(void)
{
  if(cache_seg)
  {
    while(free_caches(0, 0));
    saved_cache_seg = cache_seg;
    cache_seg = 0;
  }
}

void enable_caches(void)
{
  if(!cache_seg)
  {
    cache_seg = saved_cache_seg;
    allocate_more_caches();
  }
}

unsigned memory_kbs(void);
#pragma aux memory_kbs = \
  "int 0x12" \
  value [ax];

// free last cache entry, and if it is modified, move it to a lower entry
static int free_last_cache(void)
{
  struct cache_entry * const cache_table = DISK_CACHE;
  unsigned n = cache_size-1;

  if(!cache_table[n].modified)
  {
    cache_size = n;
    cache_freed:
    delete_alloc(cache_table[cache_size].buff_seg);
    return 1;
  }
  else while(n--)
  {
    if(!cache_table[n].modified)
    {
      cache_size--;
      nmemcpy(&cache_table[n], &cache_table[cache_size], sizeof(struct cache_entry));
      goto cache_freed;
    }
  }
  return 0;
}

int free_caches(unsigned const prog, int const dontwrite)
{
  struct cache_entry *const cache_table = DISK_CACHE;
  unsigned char rc;

  if(rc = free_usercache())
    return rc & 1;
  if(!cache_size) return 0;

  if(prog > 1)
    if(cache_table[cache_size-1].buff_seg >= (memory_kbs()<<6)-(KERNEL_ONLY_MEM>>4))
      return 0;

  if(WRITE_FLAG)
    if(!dontwrite && !DONT_FLUSH_CACHES)
      flush_caches(cache_size - 1);
    else
      return free_last_cache();

  cache_size--;

  delete_alloc(cache_table[cache_size].buff_seg);

  return 1;
}

void allocate_more_caches(void)
{
  register unsigned n;
  unsigned char save_error_code;
  struct cache_entry *const cache_table = DISK_CACHE;
  unsigned char old_strategy = get_memory_allocation_strategy();

  if(!cache_seg) return;

  save_error_code = error_code;

  set_memory_allocation_strategy(MEMALLOC_STRATEGY_LAST | MEMALLOC_STRATEGY_KEEPTOPFREE);

  for(n=cache_size;n<CACHE_SIZE;n++)
  {
    if(free_paras(0, 0, MAX_SECTOR_SIZE>>4) >= (MAX_SECTOR_SIZE>>4))
    {
      cache_table[n].buff_seg = new_alloc(MAX_SECTOR_SIZE>>4, 0);
      cache_table[n].valid = 0;
      set_mcb_type(cache_table[n].buff_seg, MCB_CACHE);
      cache_size++;
    }
    else break;
  }

  set_memory_allocation_strategy(old_strategy);

  error_code = save_error_code;
}

static void sector_in_cache(lbaindex_t const sector, unsigned char const drive, unsigned const count, struct cache_retstr * const retstr)
{
  register int n;
  struct cache_entry * const cache_table = DISK_CACHE;
  unsigned char index;
  unsigned i;
  unsigned char const max_raw_sectors = DRIVE_MIN_RAWBUFF_SECTORS(drive);
  lbaindex_t cached_sector;
#ifdef USE_32BIT
  lbaindex_t const max_sector = sector;
  b32_lba_adds(&max_sector, &count);
#else
  lbaindex_t const max_sector = sector + count;
#endif

  retstr->index = -1;

  nmemset(retstr->found, 0, max_raw_sectors);

  for(n=cache_size;n;)
  {
    n--;
    if(cache_table[n].valid)
    {
       if(cache_table[n].drive == drive)
       {
         if((cached_sector = cache_table[n].sector) == sector)
         {
           retstr->index = n;
           goto addrefs;
         }
#ifdef USE_32BIT_COMPARE
         if(b32_llgt(&cached_sector, &sector) && b32_llgt(&max_sector, &cached_sector))
         {
#else
         if(cached_sector > sector && cached_sector < max_sector)
         {
#endif
           if((index = (unsigned char)cached_sector - (unsigned char)sector) < max_raw_sectors)
             retstr->found[index]--;
           addrefs:if((i = cache_table[n].refs + ADDREFS) > cache_table[n].refs)
             cache_table[n].refs = i;
           cache_table[n].norefs = 0;
           continue;
         }
       }
       if(cache_table[n].refs)
         cache_table[n].refs--;
       cache_table[n].norefs++;
    }
  }
}

static void read_cache(lbaindex_t const sector, void __far * const buff, unsigned char const drive, unsigned const count, struct cache_retstr * const retstr)
{
  struct cache_entry * const cache_table = DISK_CACHE;

  sector_in_cache(sector, drive, count, retstr);
  if(retstr->index != -1)
    memcpy(buff, MK_FP(cache_table[retstr->index].buff_seg, 0), cache_table[retstr->index].length);
}

static int cache_find_least_popular(unsigned no_flush)
{
  register int n;
  unsigned lp = -1;
  unsigned lpn = -1;
  unsigned lpm = -1;
  unsigned max_norefs = cache_size << 4;
  struct cache_entry * const cache_table = DISK_CACHE;

  for(n=cache_size;n;)
  {
    n--;
    if(!cache_table[n].valid || cache_table[n].norefs > max_norefs)
      return n;
    if(cache_table[n].refs < lp)
    {
      lp = cache_table[n].refs;
      if(cache_table[n].modified)
        lpm = n;
      else
        lpn = n;
    }
  }
  if(lpn != -1 || no_flush)
    return lpn;
  return lpm;
}

static int save_to_cache(lbaindex_t const sector, unsigned char const drive, unsigned const count, void __far * const buff, unsigned const refs)
{
  struct cache_retstr cache_ret;
  struct cache_entry *const cache_table = DISK_CACHE;
  unsigned const len = DRIVE_SECTOR_SIZE(drive);
  sector_in_cache(sector, drive, count, &cache_ret);

  if(cache_ret.index == -1)
  {
    cache_ret.index = cache_find_least_popular(refs & 0x4000);
    if(cache_ret.index == -1) return -1;
    if(cache_table[cache_ret.index].modified)
      flush_caches(cache_ret.index);
    cache_table[cache_ret.index].drive = drive;
    cache_table[cache_ret.index].sector = sector;
    cache_table[cache_ret.index].length = len;
    cache_table[cache_ret.index].valid = 1;
  }
  else if(!(refs & 0x8000))
    goto addrefs;

  memcpy(MK_FP(cache_table[cache_ret.index].buff_seg, 0), buff, len);
  addrefs:cache_table[cache_ret.index].refs += refs & 0x3FFF;
  cache_table[cache_ret.index].norefs = 0;

  return cache_ret.index;
}

inline int write_to_cache(lbaindex_t const sector, void __far *const buff, unsigned char const drive)
{
  int index = save_to_cache(sector, drive, 1, buff, DEFAULT_REFS | 0x8000);
  if(index != -1)
  {
    DISK_CACHE[index].modified = 1;
    DRIVE_TABLE_ARR[drive*DRIVE_STRUCT_SIZE] |= SAVE_CACHE_BIT;
    DYNFLAGS |= WRITE_BIT;
  }
  return index;
}

static void set_rawcache_last(lbaindex_t, unsigned char, unsigned char);

void invalidate_caches(unsigned char const drive)
{
  register unsigned n = cache_size;
  struct cache_entry * const cache_table = DISK_CACHE;

  while(n)
  {
    n--;
    if(cache_table[n].drive == drive)
      cache_table[n].valid = 0;
  }

  if(RAWCACHE_LAST_DISK == drive)
    invalidate_rawcache();
}


