#include "globals.c"
#include "macros.c"
#include "structs.c"

unsigned directory_path(char __far *dst, unsigned char drive)
{
  dirindex_t cwd[MAX_DIR_DEPTH+1];
  dirindex_t idir[MAX_DIR_DEPTH+1];
  unsigned n = 0;
  char filename[13];
  int rc;
  unsigned i = 0;

  if(fs_changed(drive) == -1)
    goto end;

  getcwd(drive, cwd);

  nmemset(idir, -1, sizeof(dirindex_t) * (MAX_DIR_DEPTH + 1));

  chdir(drive, idir);

  while(idir[n] != cwd[n]) 
  {
    idir[n] = cwd[n];
    if(directory_entry(drive, cwd[n], filename, 1)
       || !(file_info(drive, cwd[n]).attr & ATTR_DIR))
      break;
    chdir(drive, idir);
    if(n) dst[i++] = '\\';
    rc = strlen(filename);
    if(i + rc >= 64) // range error
    {
      error_code = 0x0D;
      break;
    }
    strcpy(&dst[i], filename);
    i += rc;
    n++;
  }

  end:
  dst[i] = 0;

  return i;
}

int set_cwd(char __far *dir, dirindex_t *old_dir)
{
  dirindex_t new_cwd[MAX_DIR_DEPTH+1];
  unsigned n = 0;
  char __far *ptr;
  struct file_info fileinfo;
  unsigned char drive;
  char ps;

  if(dir[1] == ':')
  {
    drive = (dir[0]-1) & 0x1F;
    if(!partition_exists(drive))
    {
      error_code = 0xF;
      return 1;
    }
    dir+=2;
    goto jump_ps;
  }
  else
  {
    drive = WORKING_DRIVE;
    jump_ps:if(!is_path_separator(*dir))
      n = getcwd(drive, new_cwd) - 1;
    else
      do
        dir++;
      while(is_path_separator(*dir));
  }

  do
  {
    new_cwd[n] = -1;
    chdir(drive, new_cwd);
    if(n >= MAX_DIR_DEPTH)
    {
      error_code = 0x05;
      break;
    }
    if(!*dir) break;
    if(*dir == '.')
    {
      if(dir[1] == '.')
      {
        if(is_path_separator(dir[2]) || !dir[2])
        {
          if(n)
            n--;
          dir+=2 + !!dir[2];
          continue;
        }
      }
      else if(is_path_separator(dir[1]) || !dir[1])
      {
        dir+=1 + !!dir[1];
        continue;
      }
    }
    ptr = find_path_separator(dir);
    if(ptr)
    {
      ps = *ptr;
      *ptr = 0;
    }
    new_cwd[n] = get_file_index(drive, dir, 0);
    if(error_code)
      return 1;
    dir += strlen(dir);
    if(ptr)
    {
      *dir++ = ps;
      while(is_path_separator(*dir))
        dir++;
    }
    fileinfo = file_info(drive, new_cwd[n]);
    if(!(fileinfo.attr & ATTR_DIR))
    {
      error_code = 0x03;
      return 1;
    }
    if(fileinfo.attr & ATTR_SYMLINK)
    {
      n = get_symlink_path(new_cwd[n], new_cwd, fileinfo.attr, 0, old_dir);
      if(n == -1)
        return 1;
      drive = WORKING_DRIVE;
      n--;
    }
    n++;
  }
  while(1);

  return 0;
}
