#include "globals.c"
#include "macros.c"
#include "structs.c"

char __based(__segname("_CODE")) partition[] = "DRV:";

#define MAXDEVNUMLEN 5

unsigned strtonum(char __far *str)
{
  unsigned ret;
  unsigned l = strlen(str);
  unsigned n, m;

  if(l > MAXDEVNUMLEN) return -1;

  n = l;
  m = 1;
  ret = 0;

  while(n--)
  {
    if(str[n] < '0' || str[n] > '9')
      return -1;
    ret += (str[n] - '0') * m;
    m *= 10;
  }

  return ret;
}

// returns 0 if filename is not a reserved filename.
// returns -1 if the filename is a partition
// else returns the segment of a driver struct
unsigned is_driver_name(char __far *filename, unsigned *retdsk, unsigned *devnum)
{
  struct driver __far *driver_ptr;
  unsigned driver_seg = FIRST_DRIVER;
  unsigned l;
  unsigned num;

  if(!cistrncmp(filename, partition, sizeof(partition)-1))
  {
    *retdsk = (filename[sizeof(partition)-1] - 1) & 0x1F;
    if(!partition_exists(*retdsk))
      error_code = 0x0F;
    *devnum = 0;
    return -1;
  }

  else while(driver_ptr = MK_FP(driver_seg, 0))
  {
    if(!cistrncmp(filename, driver_ptr->name, l = strlen(driver_ptr->name)))
    {
      if(!filename[l])
        num = 0;
      else
      {
        num = strtonum(&filename[l]);
        if(num == -1) return 0;
      }
      if(driver_ptr->type)
        *devnum = -1;
      else
        *devnum = num;
      return driver_seg;
    }
    driver_seg = driver_ptr->next;
  }
  return 0;
}

void handle_dev(unsigned handle, unsigned dev, unsigned devnum)
{
  unsigned handle_seg = find_handle(CURRENT_PSP, handle);
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  struct partition __far *part_ptr;
  unsigned n;

  handle_ptr->file[0] = dev;

  // if the handle refers to a partition and is read-only, limit the length
  if((handle_ptr->type & 0x8003) == 0x8001)
  {
    part_ptr = MK_FP(PARTITION_SEG[dev], 0);
    handle_ptr->length = part_ptr->size;
    for(n=part_ptr->block_size;n>>=1;)
      handle_ptr->length <<= 1;
  }
  else
    handle_ptr->length = -1;

  if(handle_ptr->type & 0x4000)
    handle_ptr->file[1] = devnum;

}

int add_driver(char __far *name, void __far *func_table, unsigned type)
{
  unsigned driver_seg;
  struct driver __far *driver_ptr;
  unsigned n;

  driver_seg = allocate_high((sizeof(struct driver)+0xF)>>4, 0);

  if(!driver_seg) return -1;
  driver_ptr = MK_FP(driver_seg, 0);

  memcpy(driver_ptr->name, name, DRIVER_NAME_LEN);
  driver_ptr->func_table = func_table;
  driver_ptr->next = 0;
  driver_ptr->type = type;
  driver_ptr->psp = CURRENT_PSP;
  *(unsigned __far*)name = FP_OFF(call_int24h_wrapper);

  if(!FIRST_DRIVER) FIRST_DRIVER = driver_seg;
  else
  {
    n=FIRST_DRIVER;
    while(1)
    {
      driver_ptr = MK_FP(n, 0);
      if(!driver_ptr->next) break;
      n = driver_ptr->next;
    }
    driver_ptr->next = driver_seg;
  }

  return 0;
}

int unload_driver(char __far *devname)
{
  unsigned retdsk, devnum;
  unsigned driver_seg = is_driver_name(devname, &retdsk, &devnum);
  unsigned prev_driver = FIRST_DRIVER;
  unsigned next_driver;
  unsigned char c;
  struct driver __far *driver_ptr;
  struct partition __far *part_ptr;

  if(!driver_seg || driver_seg == -1)
    return -1;

  driver_ptr = MK_FP(driver_seg, 0);
  if(driver_ptr->type)
    for(c=MAX_PARTITIONS;c--;)
    {
      part_ptr = MK_FP(PARTITION_SEG[c], 0);
      if(part_ptr->d.driver_seg == driver_seg)
      {
        remove_partition(c);
        fsdriver_unmount_partition(driver_seg, c);
      }
    }
  next_driver = driver_ptr->next;

  if(driver_seg != prev_driver)
  {
    while(1)
    {
      driver_ptr = MK_FP(prev_driver, 0);
      if(driver_ptr->next == driver_seg)
        break;
      prev_driver = driver_ptr->next;
    }
    driver_ptr->next = next_driver;
  }

  delete_alloc(driver_seg);

  return 0;
}

// static partition-as-a-file driver functions...

extern unsigned char __far *buffer_segment;

unsigned write_drive(unsigned char drive, unsigned long offset, unsigned char __far *buff, unsigned count)
{
  unsigned sector_size = DRIVE_SECTOR_SIZE(part2logicaldrive(drive));
  unsigned n;
  unsigned long sector = offset;
  unsigned char __far *block_buff = buffer_segment;// = alloc_block_buff(sector_size);
  unsigned copy_count;
  unsigned read_count;
  struct partition __far *part = MK_FP(PARTITION_SEG[drive], 0);

  if(part->ext)
    return fsdriver_raw_write(offset, buff, count, drive);

  offset &= sector_size - 1;
  for(n=sector_size;n >>= 1;)
    sector >>= 1;

  for(n=0;;)
  {
    read_count = count / sector_size;
    if(!read_count) read_count++;
    if(offset || count < sector_size)
      if(part_read(sector, drive, read_count))
        break;
    copy_count = sector_size-offset;
    if(count < copy_count) copy_count = count;
    fix_farptr_offset((void __far **)&buff);
    memcpy(&block_buff[offset], buff, copy_count);
    if(part_write(sector, drive))
      break;
    buff += copy_count;
    n += copy_count;
    if(!(count -= copy_count))
      break;
    sector++;
    offset = 0;
  }

  return n;
}

#define DRIVER_READ_FUNC 0
#define DRIVER_WRITE_FUNC 1
#define DRIVER_SEEK_FUNC 2
#define DRIVER_OPEN_FUNC 3
#define DRIVER_CLOSE_FUNC 4
#define DRIVER_GETERROR_FUNC 5
#define DRIVER_CMD_FUNC 6
#define DRIVER_IOCTL_FUNC 7

void driver_func_entry(unsigned *old_flags, unsigned *old_psp, unsigned drv_seg)
{
  struct driver __far * const driver_ptr = MK_FP(drv_seg, 0);
  *old_flags = (INDOS_FLAG << 8) | internal_flags;
  disable();
  INDOS_FLAG = 0;
  internal_flags |= IN_KERNEL_FLAG;
  *old_psp = CURRENT_PSP;
  if(!DRVCALL_PSP)
    DRVCALL_PSP = CURRENT_PSP;
  CURRENT_PSP = driver_ptr->psp;
  enable();
}

void driver_func_ret(unsigned old_flags, unsigned old_psp)
{
  disable();
  internal_flags = old_flags;
  INDOS_FLAG = old_flags >> 8;
  CURRENT_PSP = old_psp;
  if(CURRENT_PSP == DRVCALL_PSP)
    DRVCALL_PSP = 0;
  enable();
}

static void devdrv_ret(unsigned drvseg, unsigned device,
                       unsigned old_flags, unsigned old_psp)
{
  driver_geterror(drvseg, device);
  driver_func_ret(old_flags, old_psp);
}

long call_driver_func_retl(void __far * __far *func, unsigned device, long count, unsigned origin);
#pragma aux call_driver_func_retl = \
  "call far ptr es:[di]" \
  value [dx ax] parm [es di] [bx] [dx ax] [cx];

long driver_seek(unsigned driver, unsigned device, long count, unsigned origin)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_SEEK_FUNC];
  long retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    error_code = 0x01;
  else
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    retval = call_driver_func_retl(&driver_ptr->func_table[DRIVER_SEEK_FUNC],
                                   device, count, origin);
    devdrv_ret(driver, device, old_flags, old_psp);
  }
  return retval;
}

int call_driver_func_rets(void __far * __far *func, unsigned device, void __far *buff, unsigned count);
#pragma aux call_driver_func_rets = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [bx] [dx ax] [cx];

int driver_write(unsigned driver, unsigned device, unsigned char __far *buff, unsigned count)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_WRITE_FUNC];
  int retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    error_code = 0x01;
  else
  {
    while(1)
    {
      driver_func_entry(&old_flags, &old_psp, driver);
      retval = call_driver_func_rets(&driver_ptr->func_table[DRIVER_WRITE_FUNC],
                                     device, buff, count);
      devdrv_ret(driver, device, old_flags, old_psp);
      if(!count | retval | error_code | (internal_flags & BRK_FLAG)) break;
      call_idle_handler();
    }
  }
  return retval;
}

int driver_read(unsigned driver, unsigned device, unsigned char __far *buff, unsigned count)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_READ_FUNC];
  int retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    error_code = 0x01;
  else
  {
    while(1)
    {
      driver_func_entry(&old_flags, &old_psp, driver);
      retval = call_driver_func_rets(&driver_ptr->func_table[DRIVER_READ_FUNC],
                                     device, buff, count);
      devdrv_ret(driver, device, old_flags, old_psp);
      if(!count | retval | error_code | (internal_flags & BRK_FLAG)) break;
      call_idle_handler();
    }
  }
  return retval;
}

int call_driver_open_func(void __far * __far *func_ptr, unsigned device, unsigned mode);
#pragma aux call_driver_open_func = \
  "call far ptr es:[di]" \
  parm [es di] [bx] [cx] value [ax];

int driver_open(unsigned driver, unsigned device, unsigned mode)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_OPEN_FUNC];
  int retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    return 0;
  else
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    retval = call_driver_open_func(&driver_ptr->func_table[DRIVER_OPEN_FUNC],
                                   device, mode);
    devdrv_ret(driver, device, old_flags, old_psp);
  }
  return retval;
}

int call_driver_close_func(void __far* __far *func_ptr, unsigned device);
#pragma aux call_driver_close_func = \
  "call far ptr es:[di]" \
  parm [es di] [bx] value [ax];

int driver_close(unsigned driver, unsigned device)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_CLOSE_FUNC];
  int retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    return 0;
  else
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    retval = call_driver_close_func(&driver_ptr->func_table[DRIVER_CLOSE_FUNC],
                                    device);
    devdrv_ret(driver, device, old_flags, old_psp);
  }
  return retval;
}

void driver_geterror(unsigned driver, unsigned device)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_GETERROR_FUNC];
  unsigned old_flags, old_psp;

  if(func_ptr)
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    error_code = call_driver_close_func(&driver_ptr->func_table[DRIVER_GETERROR_FUNC],
                                        device);
    driver_func_ret(old_flags, old_psp);
  }
  if(error_code > 0x12)
  {
    extended_error_code = error_code;
    error_code = 0x0E;
  }
}

int call_driver_cmd_func(void __far * __far *func_ptr, unsigned device, void __far *cmd);
#pragma aux call_driver_cmd_func = \
  "call far ptr es:[di]" \
  parm [es di] [bx] [dx ax] value [ax];

int driver_cmd(unsigned driver, unsigned device, void __far *cmd)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_CMD_FUNC];
  int retval;
  unsigned  old_flags, old_psp;

  if(!func_ptr)
    error_code = 0x01;
  else
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    retval = call_driver_cmd_func(&driver_ptr->func_table[DRIVER_CMD_FUNC],
                                  device, cmd);
    devdrv_ret(driver, device, old_flags, old_psp);
  }
  return retval;
}

unsigned long call_driver_ioctl_func(void __far * __far *func_ptr, unsigned device,
                                     unsigned bytes, unsigned char __far* buff,
                                     unsigned function);
#pragma aux call_driver_ioctl_func = \
  "call far ptr es:[di]" \
  value [dx ax] parm [es di] [bx] [cx] [dx ax] [si];

unsigned long driver_ioctl(unsigned driver, unsigned device, unsigned bytes,
                           unsigned char __far *buff, unsigned function)
{
  struct driver __far *driver_ptr = MK_FP(driver, 0);
  void __far *func_ptr = driver_ptr->func_table[DRIVER_IOCTL_FUNC];
  unsigned long retval;
  unsigned old_flags, old_psp;

  if(!func_ptr)
    error_code = 0x01;
  else
  {
    driver_func_entry(&old_flags, &old_psp, driver);
    retval = call_driver_ioctl_func(&driver_ptr->func_table[DRIVER_IOCTL_FUNC],
                                    device, bytes, buff, function);
    devdrv_ret(driver, device, old_flags, old_psp);
  }
  return retval;
}

unsigned read_drive(unsigned char drive, unsigned long offset, unsigned char __far *buff, unsigned count)
{
  unsigned sector_size = DRIVE_SECTOR_SIZE(part2logicaldrive(drive));
  unsigned n;
  unsigned long sector = offset;
  unsigned char __far *block_buff = buffer_segment;
  unsigned copy_count;
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[drive], 0);
  unsigned max_sectors, read_count;

  if(part_ptr->ext)
    return fsdriver_raw_read(offset, buff, count, drive);

  offset &= sector_size - 1;
  for(n=sector_size;n >>= 1;)
    sector >>= 1;

  for(n=0;sector < part_ptr->size;)
  {
    read_count = count/sector_size;
    if(!read_count) read_count++;
    max_sectors = part_ptr->size - sector;
    if(read_count > max_sectors) read_count = max_sectors;
    if(part_read(sector, drive, read_count))
      break;
    copy_count = sector_size-offset;
    if(count < copy_count) copy_count = count;
    fix_farptr_offset((void __far**)&buff);
    memcpy(buff, &block_buff[offset], copy_count);
    buff += copy_count;
    n += copy_count;
    if(!(count -= copy_count))
      break;
    sector++;
    offset = 0;
  }

  return n;
}
