
#include "globals.c"
#include "macros.c"
#include "structs.c"

char __based(__segname("_CODE")) loading[] = "\r\nLoading ST-DOS kernel 0.1 ";

char __based(__segname("_CODE")) fromfloppy[] = "from floppy!\r\n";
char __based(__segname("_CODE")) fromhd[] = "from hard disk!\r\n";

char __based(__segname("_CODE")) initmcb[] = "Initialized MCB.\r\n";

char __based(__segname("_CODE")) licenseinfo[] = 
"ST-DOS v0.9a  Copyright (C) 2024  Sami Tikkanen\r\n\
This program comes with ABSOLUTELY NO WARRANTY; for details see LICENSE.TXT.\r\n\
This is free software, and you are welcome to redistribute it under certain\r\n\
conditions; type `type license.txt' in the command prompt for details.\r\n";

char __based(__segname("_CODE")) stackpadding[512];

unsigned memory_kbs(void);
#pragma aux memory_kbs = \
  "int 0x12" \
  value [ax];

/*
  This function is used to do the first memory allocation. Returns the segment of the allocated memory block.
  Do not call any other function from this function. The stack is very small

  MEMORY MAP:
  0060:0000: (with the default bootloader) KERNEL
  0060:KRNLLEN: DISK I/O BUFFER
  0060:KRNLLEN+(RAW_CACHES<<4): FIRST MCB + KERNEL HEAP
 * */
inline void first_alloc(void)
{
  unsigned segment = CODE_SEGMENT+((FP_OFF(code_end)+0xF)>>4);
  unsigned rawcache_segs;

  // use smaller raw cache if the computer has only 256 kB or less memory
  if(memory_kbs() > 256) rawcache_segs = RAW_CACHES;
  else
    rawcache_segs = MAX_SECTOR_SIZE>>4;

  segment += FIRST_ALLOC_PARAGRAPHS;

  initial_mcb = segment;
  NEXT_MCB(segment) = 0;
  PREV_MCB(segment) = 0;
  OWNER_PSP(segment) = 0;
  MCB_TYPE(segment) = MCB_SYSTEM;
  PARAGRAPH_COUNT(segment) = rawcache_segs;

  nmemset(code_end, 0, FIRST_ALLOC_PARAGRAPHS<<4);
  LAST_MCB = initial_mcb;
  FILE_BUFFER_SIZE = DEFAULT_BUFFER_SIZE;

  rawcache_buff = (void *)(((segment + MCB_PARAGRAPHS) - CODE_SEGMENT) << 4);
  rawcache_size = rawcache_segs << 4;

  putstr(initmcb);
}

void init_stack(void);

void ret(void);
#pragma aux ret = \
  "jmp near ptr init_stack";

void init_kernel_1(void);
#pragma aux init_kernel_1 aborts;
void init_kernel_1(void)
{
  unsigned char source_drive_type = *(unsigned char __far*)MK_FP(0, 0xFC15);
  unsigned long source_partition_offset = *(unsigned long __far*)MK_FP(0, 0xFC1C);

  putstr(loading);

  if(!(source_drive_type & 0x80))
    putstr(fromfloppy);
  else
    putstr(fromhd);

  putstr(licenseinfo);

  first_alloc();

  // temporarily store the offset somewhere...
  BOOT_PARTITION_OFFSET = source_partition_offset;

  SYSTEM_DRIVE = source_drive_type;

  INT15_BUSY_DEVICE = -1;

//  KERNEL_SS = CODE_SEGMENT;

  ret();
}


