#include "globals.c"
#include "macros.c"
#include "structs.c"

char __based(__segname("_CODE")) unknownfn[] = "Syscall 44: Unknown function ";

static struct file_handle __far *get_handle_ptr(unsigned handle)
{
  unsigned handle_seg = find_handle(CURRENT_PSP, handle);
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  
  if(!handle_seg)
    return 0;
  
  return handle_ptr;
}

#define NOTREADY_FLAG 0x0200

#define EOF_FLAG 0x0040
#define ISDEV_FLAG 0x0080
#define CTRL_FLAG 0x4000
#define BINARY_FLAG 0x0020
#define ISCLK_FLAG 0x0008
#define ISNUL_FLAG 0x0004
#define ISCOT_FLAG 0x0002
#define ISCIN_FLAG 0x0001
#define IS_WRITTEN_FLAG EOF_FLAG

// get input status
// return 0 if EOF, else return -1
inline unsigned char ioctl6(struct file_handle __far *handle_ptr)
{
  if(handle_ptr->offset == handle_ptr->length)
    return 0;
  return -1;
}

// get output status
// 0 == not ready, -1 == ready
inline unsigned char ioctl7(struct file_handle __far *handle_ptr)
{
  return -1;
}

inline unsigned long ioctl0(struct file_handle __far *handle_ptr)
{
  unsigned retval = 0;

  if(!FP_SEG(handle_ptr)) return 0;

  if(!handle_ptr->type) // console device
  //      return 0x80D3; // this is what FreeDOS kernel returns
//      return 0x4893; // supports IOCTL, not removable,
                     // character device & uses dos int 29,
                     // standard input device & standard output device

      return 0x0883; // doesn't support dos int 29 && ioctl 2 and 3
  // else if file
  else if(!handle_ptr->is_written) retval = IS_WRITTEN_FLAG;

  return retval | handle_ptr->file[0];
}

/*inline unsigned long ioctl1(struct file_handle __far *handle_ptr, unsigned data)
{
  if(!FP_SEG(handle_ptr)) return 0;  
  
  switch(handle_ptr->type)
  {
    case 0:
      return data & ~0x40; // do nothing for now
  }
}*/

unsigned long int21c44(unsigned function, unsigned handle_dev, unsigned bytes, unsigned char __far *data);
#pragma aux int21c44 value [dx ax] parm [ax] [bx] [cx] [es dx] modify exact [];

unsigned long int21c44(unsigned function, unsigned handle_dev, unsigned bytes, unsigned char __far *data)
{
  struct file_handle __far *handle_ptr;
//  unsigned char subfunction;
  function = XCHG(function);
//  subfunction = function;
#define subfunction (function & 0xFF)
  if(subfunction != 4 && subfunction != 5 && subfunction != 8 && subfunction != 0xD)
  {
    handle_ptr = get_handle_ptr(handle_dev);
    if(FP_SEG(handle_ptr))
    {
      if(handle_ptr->type & 0x4000)
        return driver_ioctl(handle_ptr->file[0], handle_ptr->file[1], bytes, data, function);
      switch(subfunction)
      {
        case 0: case 1:
          return ioctl0(handle_ptr) << 16;
        case 6:
          return ioctl6(handle_ptr) | subfunction | ((unsigned long)(data)<<16);
        case 7:
          return ioctl7(handle_ptr) | subfunction | ((unsigned long)(data)<<16);
//        case 1:
//          return ioctl1(handle_dev, (unsigned)data) << 16;
        default:
          putstr(unknownfn);
          putstr(hex2str(subfunction));
          error_code = 0x01;
      }
    }
  }
  return (unsigned long)(data)<<16;
}
