.8086

.MODEL small

public int21h_handler_
public call_int24h_wrapper_
public int28h_handler_
public int23h_terminate_
public int20h_end
extern hex2str_:BYTE
extern putstr_:BYTE
;extern _int21h_service:BYTE
extern _int21h_function:BYTE
;extern int21awr_:BYTE
extern _code_end:BYTE
extern _error_code:BYTE

extern flush_caches_:BYTE
extern print_sp_:BYTE
extern print_regs_:BYTE
;extern _internal_flags:BYTE
_internal_flags EQU _code_end+57
;extern check_break_key_:BYTE
extern get_vector_:BYTE
extern memory_integrity_check_:BYTE

_TEXT               SEGMENT BYTE PUBLIC USE16 'CODE'
                ASSUME CS:_TEXT                               

breakstr: db "^C", 0

int21h_handler_ proc near
  test cs:_internal_flags, 0x01     ; is ctrl+break pressed?
  jz no_break
  and cs:_internal_flags, NOT 0x01  ; unset BREAK flag
  or cs:_internal_flags, 0x08       ; set BREAK_DETECTED flag

  push ax
  xor ax, ax
  int 0x16                          ; read the keypress from buffer
  pop ax

  push es
  push si

  mov si, cs
  mov es, si

  mov si, offset breakstr
  call near ptr putstr_             ; print "^C"

  pop si
  pop es

  test cs:_internal_flags, 0x04     ; is BREAK sensing enabled?
  jz no_break                       ; if not, go to break handler only from
                                    ; file I/O syscalls

  and cs:_internal_flags, NOT 0x08  ; unset BREAK_DETECTED

  mov word ptr cs:[_code_end+28], sp ; save SP
  int 0x23
  cmp word ptr cs:[_code_end+28], sp    ; did the break handler return with iret?
  je no_break                       ; if yes, jump to no_break

  mov sp, word ptr cs:[_code_end+28]  ; else restore sp
  jnc no_break                      ; and if carry is set
  mov ax, 0x4C00                    ; terminate the program

  no_break:

;    call near ptr print_sp_
;    call near ptr print_regs_

  push ds
  push bp
  push es

  mov bp, sp
  mov es, [bp+10]                ; return flags to ES register

  ; if debugging, set trap flag
IFDEF debug
  pushf
  or word ptr [bp-2], 0x0100
  popf
ENDIF

  mov word ptr cs:[_code_end+34], es       ; save flags

  test byte ptr cs:[_code_end+27], 0xFF ; is IN_DOS flag already set?
  jz set_dos_flag
;  mov cs:_error_code, 0x05       ; don`t set error code
  mov ax, 0x05                   ; return error code (access denied) to ax
  jmp access_denied

  set_dos_flag:
  mov byte ptr cs:[_code_end+27], 1 ; set IN_DOS flag

  test cs:_internal_flags, 0x02 ; are we already in kernel stack?
  jnz in_kernel_1

  mov es, word ptr cs:[_code_end+4] ; program PSP to ES
  mov es:[0x28], sp    ; save program`s stack
  mov es:[0x30], ss    ; 
  pop es               ; 
  push cs ; load kernel stack
  pop ss
  mov sp, word ptr cs:[_code_end+36]
  mov bp, es           ; original value of es to bp

  in_kernel_1:
  sti

  push ds ; es = ds
  pop es

IFDEF USE_32BIT
.386
  push eax
  push ebx
  push edx
ENDIF
IFDEF __386__
.386
  push fs
  push gs
ENDIF

;  push sp        ; for 8086
  push bp         ; fix for 80286+
  mov bp, sp      ;
  xchg bp, ss:[bp]   ; /

  push si    ; push SI
  push bp    ; and BP (original value of ES)

  xchg ah, al
  mov si, ax
  and si, 0x00FF
  shl si, 1  ; si = offset in jump table

  mov bp, cs
  mov ds, bp                 ; ds = kernel cs

  mov  ds:_error_code, 0 ; zero error code


;call near ptr print_sp_
;call near ptr print_regs_

  ; original contents of AL are in AH
  ; original contents of DS are in ES
  ; original contents of ES are in [sp]
  ; original contents of SI are in [sp + 2]
  ; [sp+4] is a pointer that points to itself
  call near ptr ds:[_int21h_function + si]

;call near ptr print_regs_
;call near ptr print_sp_

  test byte ptr [_code_end+40], 0x02
  jz no_cache_flush
  test _internal_flags, 0x02
  jnz no_cache_flush
  push ax
  mov ax, -1
  call near ptr flush_caches_
  pop ax
  no_cache_flush:

;  push ax
;  call near ptr memory_integrity_check_
;  test ax, 0xFFFF
;  jz memory_ok
;  cli
;  hlt
;  memory_ok:
;  pop ax

  pop es
  pop si ; pop bp (es) and si from the kernel stack

  add sp, 2

IFDEF __386__
.386
  pop gs
  pop fs
ENDIF
IFDEF USE_32BIT
.386
  mov bp, sp
  push dx
  push bx
  push ax

  mov edx, [bp]
  mov ebx, [bp+4]
  mov eax, [bp+8]
 
  pop ax
  pop bx
  pop dx

  add sp, 12
ENDIF

  test _internal_flags, 0x02
  jnz in_kernel_3

  int20h_end:
  mov ds, word ptr [_code_end+4]    ; program PSP to DS
  mov ss, word ptr ds:[0x30]  ; restore program stack
  mov sp, word ptr ds:[0x28]
  push cs
  pop ds

  in_kernel_3:

  cli
  mov byte ptr [_code_end+27], 0  ; unset IN_DOS flag

  test _error_code, 0xFF
  jz int21_nocarry

  mov al, _error_code
  xor ah, ah
  cmp al, 0x13
  jc no_extended_error_code
  mov word ptr [_code_end+52], ax
  mov al, 0x0E
  no_extended_error_code:
  set_carry:
  or word ptr [_code_end+34], 0x0001  ; set carry flag
;  call near ptr print_regs_
  jmp int21_end

  int21_nocarry:
  and word ptr [_code_end+34], 0xFFFE ; unset carry flag

  int21_end:

  add sp, 2

  push si
  mov si, word ptr [_code_end+34] ; other return flags
  mov bp, sp
  mov word ptr ss:[bp+10], si
  pop si

  pop bp
  pop ds

;    call near ptr print_regs_
;    call near ptr print_sp_

  int28h_handler_: ; point to this same iret instruction...
  iret

  access_denied:
  pop es
  sub sp, 2

  push cs
  pop ds
  jmp set_carry
int21h_handler_ endp

int23h_terminate_ proc near
  mov ax, 0x4C00
  jmp no_break
int23h_terminate_ endp

call_int24h_wrapper_ proc far
IFDEF __386__
  push fs
  push gs
ENDIF
  ; BP MUST be saved because the int24h handler doesnt have to preserve it
  push bp
  push es
  push cx
  push dx
  push [_code_end+4]  ; save CURRENT_PSP to stack
  push [_code_end+36] ; save KERNEL_SP to stack

  ; set the saved KERNEL_SP to its current value
  mov word ptr [_code_end+36], sp

  mov cx, word ptr [_code_end+38] ; are we in a syscall and a driver called
                                  ; this function?
  jcxz test_program_int24h
             
  mov word ptr [_code_end+4], cx   ; if yes, set CURRENT_PSP to the program
  jmp switch2userstack_int24h

  test_program_int24h:
  mov cx, word ptr [_code_end+4]
                                  ; are we running a program?
                                  ; if not, there is no user stack
  jcxz call_int24h
  
  switch2userstack_int24h:
  or byte ptr [_code_end+40], 0x08  ; set IN_CRITERR_HDL
  mov es, cx  
  mov ss, es:[0x30]                 ; switch to user stack
  mov sp, es:[0x28]

  ; the int21h handler will write wrong values to PSP, so make backups of them
  push es:[0x30]
  push es:[0x28]

  call_int24h:
  xor dl, dl
  cli
  xchg dl, byte ptr [_code_end+27] ; clear IN_DOS flag
  int 0x24
  mov byte ptr [_code_end+27], dl  ; restore IN_DOS flag
  sti

  jcxz int24_restore_kernelstack ; are we running a program?

  ; restore correct program ss:sp
  mov es, cx
  pop es:[0x28]
  pop es:[0x30]
  and byte ptr [_code_end+40], NOT 0x08

  int24_restore_kernelstack:
  mov dx, cs ; restore kernel stack
  mov ss, dx
  mov sp, word ptr [_code_end+36]

  ; restore saved KERNEL_SP
  pop [_code_end+36]

  pop [_code_end+4]
  pop dx
  pop cx
  pop es
  pop bp
IFDEF __386__
  pop gs
  pop fs
ENDIF
  int24h_return:
retf
call_int24h_wrapper_ endp

_TEXT ENDS

END
