#include "globals.c"
#include "macros.c"
#include "structs.c"

void set_memory_allocation_strategy(unsigned char strategy)
{
  internal_flags &= ~0x70;
  internal_flags |= strategy;
}

unsigned char get_memory_allocation_strategy(void)
{
  return internal_flags & 0x70;
}

// return memory size in bytes / 16
unsigned memory_size(void);
#ifdef __286__
#pragma aux memory_size = \
  "int 0x12" \
  "shl ax, 6" \
  value [ax];
#else
#pragma aux memory_size = \
  "int 0x12" \
  "push cx" \
  "mov cl, 6" \
  "shl ax, cl" \
  "pop cx" \
  value [ax];
#endif

#define MEMORY_SIZE memory_size()

static unsigned unused_paragraphs_after_mcb(unsigned mcb, unsigned owner_psp)
{
  struct mcb __far *mcb_ptr = MK_FP(mcb, 0);
  unsigned next_free_paragraph = mcb + MCB_PARAGRAPHS + mcb_ptr->paragraph_count;
  unsigned total_paragraphs = MEMORY_SIZE;

  if(owner_psp & 0x02)
  {
    while((MCB_TYPE(mcb_ptr->next_mcb_seg)&0x7F) == MCB_CACHE)
      mcb_ptr = MK_FP(mcb_ptr->next_mcb_seg, 0);
  }

  // the last KB is for the kernel
  if(owner_psp & 0x01)
  {
    total_paragraphs -= KERNEL_ONLY_MEM>>4;
    if(next_free_paragraph >= total_paragraphs)
      return 0;
  }

  if(!mcb_ptr->next_mcb_seg || mcb_ptr->next_mcb_seg > total_paragraphs)
    return total_paragraphs - next_free_paragraph;
  else
    return mcb_ptr->next_mcb_seg - next_free_paragraph;
}



unsigned new_mcb(unsigned prev_mcb, unsigned size, unsigned owner_psp)
{
  unsigned segment;
  unsigned top;
  unsigned max_size;
  unsigned raw_size = size + MCB_PARAGRAPHS;

  if(internal_flags & MEMALLOC_STRATEGY_LAST)
  {
    if(!(top = NEXT_MCB(prev_mcb)))
      top = MEMORY_SIZE;
    if(internal_flags & MEMALLOC_STRATEGY_KEEPTOPFREE)
    {
      max_size = unused_paragraphs_after_mcb(prev_mcb, !!owner_psp);
      if(max_size - MEMALLOC_TOPSIZE >= raw_size)
        NEXT_MCB(prev_mcb) = top - (MEMALLOC_TOPSIZE + raw_size);
      else
        goto firstocc;
    }
    else
      NEXT_MCB(prev_mcb) = top - raw_size;
  }
  else
    firstocc:NEXT_MCB(prev_mcb) = PARAGRAPH_COUNT(prev_mcb) + MCB_PARAGRAPHS + prev_mcb;

  segment = NEXT_MCB(prev_mcb);
  
  PARAGRAPH_COUNT(segment) = size;
  OWNER_PSP(segment) = owner_psp;
  PREV_MCB(segment) = prev_mcb;
  MCB_TYPE(segment) = 0;
  
  return segment;
}

void set_mcb_type(unsigned mcb, unsigned type)
{
  struct mcb __far *mcb_ptr = MK_FP(mcb-MCB_PARAGRAPHS, 0);

  mcb_ptr->type = type;
}

void set_mcb_owner(unsigned mcb, unsigned psp)
{
  struct mcb __far *mcb_ptr = MK_FP(mcb-MCB_PARAGRAPHS, 0);

  mcb_ptr->owner_psp = psp;
}

int defrag_memory(void);

unsigned make_more_memory(unsigned req)
{
  free_old_dta_structs();
  while(free_caches(0, 0));
  defrag_memory();
  allocate_more_caches();
  return free_paras(1, 1, req);
}

static int alloc_cond(unsigned psp, unsigned *t)
{
  if(INT15_BUSY_DEVICE == 0xFF)
  {
    if(flush_buffers() || free_caches(psp, 0))
      return 1;
    else if((*t)--)
      return !defrag_memory();
    else return 0;
  }
  else
    return free_caches(psp, 1);
}

/*
 * Returns the segment of the allocated memory block, or 0 if out of memory.
 * */
unsigned new_alloc(unsigned paragraphs, unsigned owner_psp)
{
  unsigned segment;
  unsigned next_segment;
  unsigned t = 1;

  if(paragraphs + MCB_PARAGRAPHS < paragraphs)
    paragraphs = -1 - MCB_PARAGRAPHS;

  free_old_dta_structs();

  do 
  {
    if(internal_flags & MEMALLOC_STRATEGY_LAST)
    {
      segment = LAST_MCB;
      next_segment = 0;
      do
      {
        if(unused_paragraphs_after_mcb(segment, !!owner_psp) >= paragraphs + MCB_PARAGRAPHS)
          goto newmcb;
        next_segment = segment;
        segment = PREV_MCB(segment);
      }
      while(segment);
    }
    else
    {
      segment = initial_mcb;
      do
      {
        next_segment = NEXT_MCB(segment);
        if(unused_paragraphs_after_mcb(segment, !!owner_psp) >= paragraphs + MCB_PARAGRAPHS)
        {
          // insert new mcb between segment and its next mcb
          newmcb:segment = new_mcb(segment, paragraphs, owner_psp);
          if(!(NEXT_MCB(segment) = next_segment))
            LAST_MCB = segment;
          else
            PREV_MCB(next_segment) = segment;
          return segment+MCB_PARAGRAPHS;
        }
        segment = next_segment;
      }
      while(segment);
    }
  }
  while(alloc_cond(owner_psp, (unsigned *)&t));

  AVAILABLE_MEMORY = free_paras(1, 1, 0);

  allocate_more_caches();
 
  error_code = 0x08;  
  return 0;  
}

unsigned allocate_high(unsigned paragraphs, unsigned owner_psp)
{
  unsigned char old_strategy = get_memory_allocation_strategy();
  unsigned retval;
  set_memory_allocation_strategy(MEMALLOC_STRATEGY_LAST);
  retval = new_alloc(paragraphs, owner_psp);
  set_memory_allocation_strategy(old_strategy);
  return retval;
}

inline int is_dma_safe(unsigned mcb, unsigned paragraphs)
{
  unsigned block_max_size = 0x1000 - ((mcb+MCB_PARAGRAPHS) & 0xFFF);
  if(paragraphs > block_max_size) return 0;
  return 1;
}

// allocate a DMA-safe memory block
/*unsigned dma_alloc(unsigned paragraphs, unsigned owner_psp)
{
  unsigned segment;
  unsigned next_segment;
  unsigned end;
  unsigned new_segment;

  do
  {
    segment = initial_mcb;
    do
    {
      next_segment = NEXT_MCB(segment);
      if(unused_paragraphs_after_mcb(segment, !!owner_psp) >= paragraphs + MCB_PARAGRAPHS)
      {
        new_segment = segment+MCB_PARAGRAPHS+PARAGRAPH_COUNT(segment);
        while(!is_dma_safe(new_segment, paragraphs))
        {
          new_segment++;
          end = new_segment+MCB_PARAGRAPHS+paragraphs;
          if(next_segment)
          {
            if(end >= next_segment)
              goto next;
          }
          else
            if(end >= MEMORY_SIZE)
              goto next;
        }
        if(!(NEXT_MCB(new_segment) = next_segment))
          LAST_MCB = new_segment;
        else
          PREV_MCB(NEXT_MCB(new_segment)) = new_segment;
        NEXT_MCB(segment) = new_segment;
        PREV_MCB(new_segment) = segment;
        PARAGRAPH_COUNT(new_segment) = paragraphs;
        OWNER_PSP(new_segment) = owner_psp;
        MCB_TYPE(new_segment) = 0;

        return new_segment + MCB_PARAGRAPHS;
      }
    next:;
    }
    while(segment = next_segment);
  }
  while(free_caches(owner_psp));

  allocate_more_caches();

  error_code = 0x08;
  return 0;
}*/

unsigned mcb_exists(unsigned mcb)
{
  unsigned segment = initial_mcb;
  
  while(segment && segment != mcb)
    segment = NEXT_MCB(segment);
  
  return segment;
}

static unsigned mcb_max_size(unsigned mcb)
{
  return PARAGRAPH_COUNT(mcb) + unused_paragraphs_after_mcb(mcb, !!OWNER_PSP(mcb));
}

unsigned free_paras(unsigned owner_psp, unsigned no_cache, unsigned req_amount)
{
  unsigned segment = initial_mcb;
  unsigned max = 0;
  unsigned temp;

  while(segment)
  {
    if((temp = unused_paragraphs_after_mcb(segment, (no_cache<<1) | !!owner_psp)) > max)
      if(req_amount && temp - MCB_PARAGRAPHS >= req_amount)
        return req_amount;
      else
        max = temp;
    segment = NEXT_MCB(segment);
  }
  if(max < MCB_PARAGRAPHS) return 0;
  return max - MCB_PARAGRAPHS;
}

void shrink_alloc(unsigned mcb, unsigned new_size)
{
  mcb -= MCB_PARAGRAPHS;
  PARAGRAPH_COUNT(mcb) = new_size;
  allocate_more_caches();
}

unsigned modify_alloc(unsigned mcb, unsigned new_size)
{
  unsigned max_size;
  unsigned retval;
  unsigned t = 1;

  mcb-=MCB_PARAGRAPHS;

  free_old_dta_structs();

  do
    if((max_size = mcb_max_size(mcb)) >= new_size)
    {
      retval = PARAGRAPH_COUNT(mcb) = new_size;
      goto ret;
    }
  while(alloc_cond(OWNER_PSP(mcb), (unsigned *)&t));

  error_code = 0x08;
  retval = PARAGRAPH_COUNT(mcb) = max_size;

  ret:
  allocate_more_caches();
  return retval;
}

void delete_mcb(unsigned mcb)
{
  unsigned next = NEXT_MCB(mcb);
  unsigned prev = PREV_MCB(mcb);

  if(!(NEXT_MCB(prev) = next))
    LAST_MCB = prev;
  else
    PREV_MCB(next) = prev;
}

void delete_alloc(unsigned mcb)
{
  delete_mcb(mcb-MCB_PARAGRAPHS);
}

unsigned free_alloc(unsigned mcb)
{
  unsigned segment = initial_mcb;

  mcb-=MCB_PARAGRAPHS;

  while(NEXT_MCB(segment) != mcb)
  {
    if(!NEXT_MCB(segment))
    {
      error_code = 0x09;
      return 1; // invalid memory block
    }
    segment = NEXT_MCB(segment);
  }

  delete_mcb(mcb);

  return 0;
}

int free_usercache(void)
{
  unsigned segment = initial_mcb;
  unsigned char type;
  struct mcb __far *mcb_ptr;

  while(segment)
  {
    mcb_ptr = MK_FP(segment, 0);
    if(((type = mcb_ptr->type) & 0x7F) == MCB_CACHE)
      if(type == MCB_USERCACHE)
      {
        if(*mcb_ptr->usercache_freedptr == 0xFE)
          goto next;
        *mcb_ptr->usercache_freedptr = -1;
        delete_mcb(segment);
        return 1;
      }
      else
        return 0;
    next:segment = mcb_ptr->next_mcb_seg;
  }
  return 2;
}

void free_psp_allocs(unsigned psp)
{
  unsigned segment = initial_mcb;

  delete_alloc(psp);
  while(segment)
  {
    if(OWNER_PSP(segment) == psp)
      delete_mcb(segment);
    segment = NEXT_MCB(segment);
  }

  allocate_more_caches();
}

char __based(__segname("_CODE")) mem_corrupted[] = "Panic: MCB corrupted\r\n";

// returns false if ok, true if memory is corrupted
int memory_integrity_check(void)
{
  unsigned segment = initial_mcb;
  unsigned next;

  if(PREV_MCB(segment) || OWNER_PSP(segment) || MCB_TYPE(segment) != MCB_SYSTEM)
    goto error;

  while(1)
  {
    if(next = NEXT_MCB(segment))
    {
      if(segment + MCB_PARAGRAPHS + PARAGRAPH_COUNT(segment) > next)
      {
        error:
        error_code = 0x07;
        putstr(mem_corrupted);
        break;
      }
      if(PREV_MCB(next) != segment)
        goto error;
    }
    else
    {
      if(segment + MCB_PARAGRAPHS + PARAGRAPH_COUNT(segment) > MEMORY_SIZE
      || segment != LAST_MCB)
        goto error;
      break;
    }
    segment = next;
  }
  return next;
}

// memory defragger functions...

static void update_dups(unsigned old_handle_seg, unsigned new_handle_seg)
{
  struct file_handle __far *handle_ptr = MK_FP(initial_mcb+MCB_PARAGRAPHS, 0);

  while(handle_ptr = MK_FP(next_handle_seg(FP_SEG(handle_ptr)), 0))
  {
    if(handle_ptr->dup == old_handle_seg)
      handle_ptr->dup = new_handle_seg;
    else if(handle_ptr->orig == old_handle_seg)
      handle_ptr->orig = new_handle_seg;
  }
}

static int move_memory_block(unsigned src_seg)
{
  struct mcb __far *mcb_ptr = MK_FP(src_seg, 0);
  void __far *dst;
  void __far *src = MK_FP(src_seg+MCB_PARAGRAPHS, 0);

  if(free_paras(0, 0, mcb_ptr->paragraph_count) != mcb_ptr->paragraph_count) return -1;
  dst = MK_FP(new_alloc(mcb_ptr->paragraph_count, 0), 0);

  memcpy(dst, src, mcb_ptr->paragraph_count << 4);

  if(mcb_ptr->type == MCB_HANDLE)
    update_dups(FP_SEG(src), FP_SEG(dst));

  set_mcb_type(FP_SEG(dst), mcb_ptr->type);

  delete_alloc(FP_SEG(src));

  return 0;
}

static int mem_block_is_movable(struct mcb __far *mcb)
{
  if(mcb->type == MCB_DTA || mcb->type == MCB_HANDLE)
    return 1;
  return 0;
}

static int move_kernel_structs_low(void)
{
  struct mcb __far *mcb;
  unsigned next = initial_mcb;
  set_memory_allocation_strategy(MEMALLOC_STRATEGY_FIRST);

  do
  {
    mcb = MK_FP(next, 0);
    next = mcb->next_mcb_seg;
    if(mem_block_is_movable(mcb))
      if(move_memory_block(FP_SEG(mcb)))
        return -1;
  }
  while(next);
  return 0;
}

static void move_kernel_structs_high(void)
{
  struct mcb __far *mcb;
  unsigned prev = LAST_MCB;
  set_memory_allocation_strategy(MEMALLOC_STRATEGY_LAST);

  do
  {
    mcb = MK_FP(prev, 0);
    prev = mcb->prev_mcb_seg;
    if(mem_block_is_movable(mcb))
      move_memory_block(FP_SEG(mcb));
  }
  while(prev);
}

int defrag_memory(void)
{
  unsigned char old_strategy = get_memory_allocation_strategy();
  int rc = move_kernel_structs_low();

  if(!rc)
    move_kernel_structs_high();

  set_memory_allocation_strategy(old_strategy);
  return rc;
}
