#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <i86.h>
#include <dos.h>

#define MEDIA_UNINITIALIZED 0xFF
#define MEDIA_FLOPPY 0
#define MEDIA_HD 8

char oem_name[6] = "ST-DOS";
char error[] = "Error";

unsigned bytes_per_sector;
unsigned char sectors_per_cluster;
unsigned reserved_sectors;
unsigned char fat_copies;
unsigned root_entries;
unsigned long total_sectors;
unsigned sectors_per_fat;
unsigned sectors_per_track;
unsigned heads;
unsigned long hidden_sectors;
unsigned tracks;

unsigned char media_descriptor = 0xF8;

unsigned media_type = MEDIA_UNINITIALIZED;

unsigned cluster_width;

unsigned char quick = 0;

int mkbootsector(FILE *fp)
{
  unsigned n;

  if(fseek(fp, 3, SEEK_SET))
    return -1;
  if(fwrite(oem_name, 6, 1, fp) != 1)
    return -1;

  if(fseek(fp, 11, SEEK_SET))
    return -1;
  if(fwrite(&bytes_per_sector, 2, 1, fp) != 1)
    return -1;

  if(fwrite(&sectors_per_cluster, 1, 1, fp) != 1)
    return -1;

  if(fwrite(&reserved_sectors, 2, 1, fp) != 1)
    return -1;

  if(fwrite(&fat_copies, 1, 1, fp) != 1)
    return -1;

  if(fwrite(&root_entries, 2, 1, fp) != 1)
    return -1;

  if(total_sectors < 0x10000)
  {
    if(fwrite(&total_sectors, 2, 1, fp) != 1)
      return -1;
  }
  else
    for(n=2;n--;)
      if(EOF == fputc(0, fp))
        return -1;

  if(fwrite(&media_descriptor, 1, 1, fp) != 1)
    return -1;

  if(fwrite(&sectors_per_fat, 2, 1, fp) != 1)
    return -1;

  if(fwrite(&sectors_per_track, 2, 1, fp) != 1)
    return -1;

  if(fwrite(&heads, 2, 1, fp) != 1)
    return -1;

  if(total_sectors > 0xFFFF)
  {
    if(fwrite(&hidden_sectors, 4, 1, fp) != 1)
      return -1;
    if(fwrite(&total_sectors, 4, 1, fp) != 1)
      return -1;
  }
  else
  {
    if(fwrite(&hidden_sectors, 2, 1, fp) != 1)
      return -1;

    if(cluster_width == 16)
      for(n=0;n<6;n++)
        fputc(0, fp);
  }

  if(fflush(fp)) return -1;

  return 0;
}

int bios_format_track(unsigned char sectors, unsigned track,
                      unsigned head_drive, void __far *);
#pragma aux bios_format_track = \
  "xchg ch, cl" \
  "ror cl, 1" \
  "ror cl, 1" \
  "mov ah, 0x05" \
  "int 0x13" \
  "jc end" \
  "xor ax, ax" \
  "end:"  \
  parm [ax] [cx] [dx] [es bx] value [ax];

unsigned char __far *afb;
unsigned char __far *old_dpt;
unsigned char temp_dpt[16];

unsigned controller_write_test(unsigned char drive, void __far *buff);
#pragma aux controller_write_test = \
  "mov ah, 0x0F" \
  "int 0x13" \
  parm [dl] [es bx] value [ax];

int format_track(unsigned char drive, unsigned track, unsigned head)
{
  unsigned n;
  unsigned interleave = 1;
  unsigned rc;
  unsigned s;
  unsigned i;

  if(sectors_per_track > 18 && sectors_per_track < 36)
  {
    interleave++;
    if(total_sectors > 1680)
      interleave++;
  }

  if(!(drive & 0x80))
  {
    for(n=0,s=1,i=0;s<=sectors_per_track;s++,n+=interleave)
    {
      if(n >= sectors_per_track)
        n = ++i;
      afb[n*4] = track;
      afb[n*4+1] = head;
      afb[n*4+2] = s;
      afb[n*4+3] = 0x02;
    }
  }
  else
  {
    if(rc = (controller_write_test(drive, afb) >> 8))
      return rc;
    for(n=0;n<sectors_per_track;n++)
    {
      afb[n*2] = 0x00;
      afb[n*2+1] = n+1;
    }
  }

  printf("\rFormatting head %u track %u  ", head, track);
  fflush(stdout);
  return bios_format_track(sectors_per_track, track, (head << 8) | drive, afb) >> 8;
}

void __far *get_dpt(unsigned drive);
#pragma aux get_dpt = \
  "mov ah, 0x08" \
  "push bx" \
  "push cx" \
  "xor di, di" \
  "mov es, di" \
  "int 0x13" \
  "jnc end" \
  "xor di, di" \
  "mov es, di" \
  "end:" \
  "pop dx" \
  "pop cx" \
  value [es di] parm [dx] modify [ax dx];

int reset_drive(unsigned char drive);
#pragma aux reset_drive = \
  "xor ah, ah" \
  "int 0x13" \
  value [ax] parm [dl];

int format_bootsector(unsigned char);

inline int set_media_type(unsigned char drive, unsigned char sectors)
{
  unsigned n;
  unsigned rc;
  old_dpt = get_dpt(drive);
  if(!old_dpt) return -1;
  _fmemcpy(temp_dpt, old_dpt, 16);
  for(temp_dpt[3] = 0, n = 128;n < bytes_per_sector;n*=2, temp_dpt[3]++);
  temp_dpt[4] = sectors_per_track;
  if(sectors_per_track > 18 && sectors_per_track < 36)
  {
    do
    {
      printf("Changing gap size from %.2X to ", temp_dpt[7]);
      if(temp_dpt[7] > 0x1C)
        temp_dpt[7] = 0x1C;
      else
        temp_dpt[7]--;
      printf("%.2X\n", temp_dpt[7]);
      _dos_setvect(0x1E, (void __far*)temp_dpt);
      if(rc = (reset_drive(drive) >> 8))
      {
        puts("Disk reset failed!");
        return rc;
      }
      puts("Trying to format track 0, head 0...");
      
      if(!(rc = format_track(drive, 0, 0)))
      {
        puts("Success, trying to write sector 1...");
        if(!(rc = format_bootsector(drive)))
        {
          puts("This gap length seems to be working for this drive.");
          temp_dpt[7] -= temp_dpt[7]/10;
          printf("Using gap length %.2X\n", temp_dpt[7]);
          return 0;
        }
      }
    }
    while(temp_dpt[7]);
    return rc;
  }
  _dos_setvect(0x1E, (void __far*)temp_dpt);
  return reset_drive(drive) >> 8;
}

int bios_write_bootsector(unsigned drive, void __far *buff);
#pragma aux bios_write_bootsector = \
  "xor ah, ah" \
  "int 0x13" \
  "mov ax, 0x0301" \
  "push cx" \
  "mov cx, 1" \
  "int 0x13" \
  "pop cx" \
  "jc end" \
  "xor ax, ax" \
  "end:"  \
  parm [dx] [es bx] value [ax];


int format_bootsector(unsigned char drive)
{
  unsigned char buff[1024];
  unsigned char *bootsector;
  unsigned dma = (FP_SEG(buff) << 4) + FP_OFF(buff);
  unsigned n;
  int rc;

  if(dma + 512 < 512) bootsector = &buff[512];
  else bootsector = buff;

  *(unsigned short*)&bootsector[11] = bytes_per_sector;
  bootsector[13] = sectors_per_cluster;
  *(unsigned short*)&bootsector[14] = reserved_sectors;
  bootsector[16] = fat_copies;
  *(unsigned short*)&bootsector[17] = root_entries;
  if(total_sectors < 0x10000)
  {
    *(unsigned short*)&bootsector[19] = total_sectors;
    *(unsigned short*)&bootsector[28] = hidden_sectors;
  }
  else
  {
    *(unsigned short*)&bootsector[19] = 0;
    *(unsigned long*)&bootsector[32] = total_sectors;
    *(unsigned long*)&bootsector[28] = hidden_sectors;
  }
  bootsector[21] = media_descriptor;
  *(unsigned short*)&bootsector[22] = sectors_per_fat;
  *(unsigned short*)&bootsector[24] = sectors_per_track;
  *(unsigned short*)&bootsector[26] = heads;
 
  puts("\nWriting boot sector values.");

  for(n=9;n--;)
   if(!(rc = bios_write_bootsector(drive, bootsector)))
     break;
  return rc >> 8;
}

int low_level_format(unsigned floppy)
{
  unsigned char __far *media_state = MK_FP(0x40, 0x90);
  unsigned char drive_type; // 0 = 360 kB 5,25", 1 = 1,2 MB 5,25"
                            // 2 = 720 kB 3,5",  3 = 1,44 MB 3,5"
  unsigned n,h,r;
  int rc;
  unsigned long target_kb_size;
  unsigned char __far *afb_alloc;
  unsigned afb_flat;
  unsigned alloc_size;

  if(floppy < 2)
    alloc_size = sectors_per_track*4;
  else
    alloc_size = 512;
  afb_alloc = malloc(alloc_size*2);

  if(!afb_alloc)
  {
    puts("Out of memory");
    return -1;
  }

  afb_flat = FP_SEG(afb_alloc) << 4;
  afb_flat += FP_OFF(afb_alloc);
  afb = afb_alloc;
  if(afb_flat + alloc_size < alloc_size)
    afb += alloc_size;

//  printf("Sector buffer: %.4X:%.4X\n", FP_SEG(afb), afb);

  printf("The drive seems to be a ");
  if(floppy >= 2)
    floppy = (floppy - 2) | 0x80;
  if(!(floppy & 0x80))
  {
    switch(media_state[floppy] & 0x7)
    {
      case 0:
      case 3:
        printf("320/360 kB 5,25\"");
        drive_type = 0;
        break;
      case 7:
        printf("3,5\"");
        if(sectors_per_track == 9 && tracks == 80)
          drive_type = 2;
        else
          drive_type = 3;
        break;
      default:
        printf("1,2 MB 5,25\"");
        drive_type = 1;
    }
  }
  else printf("hard");
  puts(" drive.");

  if(bytes_per_sector <= 1024)
  {
    target_kb_size = (unsigned long)sectors_per_track * tracks * heads;
    target_kb_size /= 1024 / bytes_per_sector;
    printf("The target disk is a ");
    if(target_kb_size < 4096)
      printf("%u kB", target_kb_size);
    else
      printf("%u MB", target_kb_size / 1024);
    puts(" disk");
  }

  if(!(floppy & 0x80))
    if(rc = set_media_type(floppy, sectors_per_track))
      return rc;

  for(n=0;n<tracks;n++)
    for(h=0;h<heads;h++)
      for(r=3;r--;)
        if(!(rc = format_track(floppy, n, h)))
          break;
        else if(!r)
          return rc;

  return format_bootsector(floppy);
}

int mkfats(FILE *fp)
{
  unsigned long l = 0;
  unsigned long t;
  unsigned n;
  unsigned f;
  unsigned char eof = -1;
  unsigned char null = 0;

  if(fseek(fp, reserved_sectors*bytes_per_sector, SEEK_SET))
    return -1;


  puts("Writing cluster tables");
  for(f=1;f<=fat_copies;f++)
  {
    if(fputc(media_descriptor, fp) != media_descriptor)
      return -1;

    if(cluster_width == 12)
      n = 2;
    else
      n = 3;

    while(n--)
      if(fputc(eof, fp) != eof)
        return -1;

    while(ftell(fp) < (reserved_sectors + (unsigned long)f*sectors_per_fat)*bytes_per_sector)
      if(fputc(null, fp) != null)
        return -1;
  }

  puts("Creating root directory...");
  for(n=32*root_entries;n--;)
    if(fputc(null, fp) != null)
      return -1;

  if(!quick)
  {
    puts("Zeroing the rest of the disk...");
    t = (unsigned long)total_sectors*bytes_per_sector;
/*    while((l = ftell(fp)) < t)
    {
      if(fputc(null, fp) != null)
        return -1;
      if(!(l & 0xFF))
        printf("\r%u%% complete", (100*l)/t);
        fflush(stdout);
    }*/
    fseek(fp, t-1, SEEK_SET);
    if(fputc(null, fp) != null)
      return -1;
    puts("\rReady.         ");
  }

  return 0;
}

int mkfs(FILE *fp)
{
  unsigned kilobytes = total_sectors >> 1;

  if(media_type == MEDIA_FLOPPY)
  {
    if(kilobytes == 320 && heads == 2 && tracks == 40 && sectors_per_track == 8)
      media_descriptor = 0xFF;
    else if(kilobytes == 160 && heads == 1 && tracks == 40 && sectors_per_track == 8)
      media_descriptor = 0xFE;
    else if(kilobytes == 360 && heads == 2 && tracks == 40 && sectors_per_track == 9)
      media_descriptor = 0xFD;
    else if(kilobytes == 180 && heads == 1 && tracks == 40 && sectors_per_track == 9)
      media_descriptor = 0xFC;
    else if(kilobytes == 640 && heads == 2 && tracks == 80 && sectors_per_track == 8)
      media_descriptor = 0xFB;
    else if(kilobytes == 320 && heads == 1 && tracks == 80 && sectors_per_track == 8)
      media_descriptor = 0xFA;
    else if((kilobytes == 1200 && heads == 2 && tracks == 80 && sectors_per_track == 15)
      || (kilobytes == 720 && heads == 2 && tracks == 80 && sectors_per_track == 9))
      media_descriptor = 0xF9;
    else// if((kilobytes == 1440 && sides == 2 && tracks == 80 && sectors_per_track == 18)
//      || (kilobytes == 2880 && sides == 2 && tracks == 80 && sectors_per_track == 36))
      media_descriptor = 0xF0;
  }
  else
  {

  }

  puts("Writing boot sector...");
  if(mkbootsector(fp))
  {
    puts(error);
    return -1;
  }
  puts("Creating filesystem...");
  if(mkfats(fp))
  {
    puts(error);
    return -1;
  }

  return 0;
}

int get_fs_parms(FILE *fp)
{
  unsigned char buf[36];

  if(fread(buf, 36, 1, fp) != 1)
    return -1;

  bytes_per_sector = *(unsigned short*)&buf[11];
  sectors_per_cluster = buf[13];
  reserved_sectors = *(unsigned short*)&buf[14];
  fat_copies = buf[16];
  root_entries = *(unsigned short*)&buf[17];
  total_sectors = *(unsigned short*)&buf[19];
  if(buf[21] == 0xF8) media_type = MEDIA_HD;
  else media_type = MEDIA_FLOPPY;
  sectors_per_fat = *(unsigned short*)&buf[22];
  sectors_per_track = *(unsigned short*)&buf[24];
  heads = *(unsigned short*)&buf[26];
  hidden_sectors = *(unsigned short*)&buf[28];

  if(!total_sectors)
  {
    total_sectors = *(unsigned long*)&buf[32];
    hidden_sectors = *(unsigned long*)&buf[28];
  }

  return 0;
}

int main(int argc, char *argv[])
{
  char *filename;
  char devname[] = "DRV:A";
  char *target;
  FILE *fp;
  unsigned size;
  int rc;
  unsigned n;

  if(argc < 2)
    goto help;

  filename = argv[1];

  if(filename[1] == ':' && strlen(filename) == 2)
  {
    if(((filename[0] - 1) & 0x1F) < 2) media_type = MEDIA_FLOPPY;
    else media_type = MEDIA_HD;
    devname[4] = filename[0];
    target = devname;
    printf("Put target disk in drive %c and press ENTER...", filename[0]);
    getchar();
  }
  else target = filename;

  if(argc == 2)
  {
    fp = fopen(target, "rb");
    if(fp)
    {
      if(get_fs_parms(fp))
      {
        noparms:
        puts("Could not detect old FS size.");
        help:
        puts("Define FS parameters by:");
        puts("> FORMAT [TARGET] [BYTES_PER_SECTOR] [SECTORS_PER_CLUSTER] [RESERVED_SECTORS] [FAT_COPIES] [ROOT_ENTRIES] [TOTAL_SECTORS] [SECTORS_PER_FAT] [SECTORS_PER_TRACK] [HEADS] [HIDDEN_SECTORS] [CLUSTER_WIDTH]");
        puts("Example for formatting a standard 1,44 MB 3,25\" diskette:");
        puts("> FORMAT A: 512 1 1 2 224 2880 9 18 2 0 12");
        puts("You can also use a pre-defined floppy disk format by using the /SIZE switch:");
        puts("> FORMAT [TARGET] /SIZE=[format]");
        puts("Supported pre-defined sizes are: 320 (320k, 2 heads, 40 tracks) 3202 (320k, 1 heads, 80 tracks) 160 360 180 640 [12]00 720 [14]40 [28]80 [168]0 [172]0");
        puts("If this fails, you probably need to low-level-format the disk first.");
        return -1;
      }
      fclose(fp);
      if(media_type == MEDIA_FLOPPY) cluster_width = 12;
      else cluster_width = 16;
      quick = 1;
    }
    else
      goto noparms;
  }
  else if(argc == 3)
  {
    if(!strncmp(argv[2], "/SIZE", 5))
    {
      if(media_type == MEDIA_UNINITIALIZED)
        media_type = MEDIA_FLOPPY;
      switch(size = atoi(&argv[2][6]))
      {
        case 320:
          heads = 2;
          sectors_per_track = 8;
          root_entries = 112;
          sectors_per_fat = 1;
          sectors_per_cluster = 2;
          tracks = 40;
          break;
        case 160:
          heads = 1;
          sectors_per_track = 8;
          root_entries = 64;
          sectors_per_fat = 1;
          sectors_per_cluster = 1;
          tracks = 40;
          break;
        case 360:
          heads = 2;
          sectors_per_track = 9;
          root_entries = 112;
          sectors_per_fat = 2;
          sectors_per_cluster = 2;
          tracks = 40;
          break;
        case 180:
          heads = 1;
          sectors_per_track = 9;
          root_entries = 64;
          sectors_per_fat = 2;
          sectors_per_cluster = 1;
          tracks = 40;
          break;
        case 640:
          heads = 2;
          sectors_per_track = 8;
          root_entries = 112;
          sectors_per_fat = 2;
          sectors_per_cluster = 2;
          tracks = 80;
          break;
        case 3202:
          heads = 1;
          sectors_per_track = 8;
          root_entries = 112;
          sectors_per_fat = 2;
          sectors_per_cluster = 1;
          tracks = 80;
          size = 320;
          break;
        case 12:
          heads = 2;
          sectors_per_track = 15;
          root_entries = 224;
          sectors_per_fat = 7;
          sectors_per_cluster = 1;
          tracks = 80;
          size = 1200;
          break;
        case 720:
          heads = 2;
          sectors_per_track = 9;
          root_entries = 112;
          sectors_per_fat = 3;
          sectors_per_cluster = 2;
          tracks = 80;
          break;
        case 14:
          heads = 2;
          sectors_per_track = 18;
          root_entries = 224;
          sectors_per_fat = 9;
          sectors_per_cluster = 1;
          tracks = 80;
          size = 1440;
          break;
        case 28:
          heads = 2;
          sectors_per_track = 36;
          root_entries = 240;
          sectors_per_fat = 9;
          sectors_per_cluster = 2;
          tracks = 80;
          size = 2880;
          break;
        case 168: // DMF1024
          heads = 2;
          sectors_per_track = 21;
          root_entries = 16;
          sectors_per_fat = 5;
          sectors_per_cluster = 2;
          tracks = 80;
          size = 1680;
        break;
        case 172:
          heads = 2;
          sectors_per_track = 21;
          root_entries = 16;
          sectors_per_fat = 6;
          sectors_per_cluster = 2;
          tracks = 82;
          size = 1720;
        break;
        default:
          puts("Unsupported format.");
          return -1;
      }
      bytes_per_sector = 512;
      total_sectors = size<<1;
      reserved_sectors = 1;
      fat_copies = 2;
      hidden_sectors = 0;
      cluster_width = 12;
    }
    else goto help;
  }
  else if(argc == 13)
  {
    if(media_type == MEDIA_UNINITIALIZED)
      media_type = MEDIA_HD;
    bytes_per_sector = atoi(argv[2]);
    sectors_per_cluster = atoi(argv[3]);
    reserved_sectors = atoi(argv[4]);
    fat_copies = atoi(argv[5]);
    root_entries = atoi(argv[6]);
//    total_sectors = atoi(argv[7]);
    total_sectors = strtoul(argv[7], 0, 10);
    sectors_per_fat = atoi(argv[8]);
    sectors_per_track = atoi(argv[9]);
    heads = atoi(argv[10]);
//    hidden_sectors = atoi(argv[11]);
    hidden_sectors = strtoul(argv[11], 0, 10);
    cluster_width = atoi(argv[12]);
    tracks = total_sectors / sectors_per_track / heads;
  }
  else goto help;
  printf("Bytes per sector: %u\nSectors per cluster: %u\nReserved sectors: %u\nFATs: %u\nRoot directory entries: %u\nTotal sectors in FS: %lu\n\
Sectors per FAT: %u\nSectors per track: %u\nHeads: %u\nHidden sectors: %lu\n", bytes_per_sector, sectors_per_cluster, reserved_sectors,
      fat_copies, root_entries, total_sectors, sectors_per_fat, sectors_per_track, heads, hidden_sectors);

  fp = fopen(target, "wb");
  if(!fp)
  {
    printf("Error opening %s\n", target);
    return -1;
  }
  if(filename[1] == ':' && !quick)
  {
    if(((filename[0]-1)&0x1F) < 2)
    {
      puts("The target seems to be a physical floppy - doing low-level format...");
      rc = low_level_format((filename[0] - 1) & 0x1F);
//      if(((filename[0] - 1) & 0x1F) < 2)
        _dos_setvect(0x1E, (void __far*)old_dpt);
      if(rc)
      {
        printf("Error %.2Xh while low level formatting\n", rc);
        return -1;
      }
    }
    quick = 1;
  }
  mkfs(fp);
  fclose(fp);

  return 0;
}
