#include <stdio.h>
#include <stdlib.h>
#include <i86.h>

#define MCB_NORMAL 0
#define MCB_DTA 1
#define MCB_CACHE 2
#define MCB_FS 3
#define MCB_HANDLE 4
#define MCB_SYSTEM 5
#define MCB_USERCACHE (MCB_CACHE|0x80)

#define MAX_DIR_DEPTH 6

struct file_handle
{
  unsigned long unused1;
  unsigned long unused2;
  unsigned long length;
  unsigned long offset;
  unsigned long file[MAX_DIR_DEPTH+3]; // drive + directory + file_index + -1
  unsigned buffer_size;
  int id;
  unsigned type;
  unsigned owner_psp;
  unsigned write_buffer;
  unsigned orig;
  unsigned dup;
};


struct mcb
{
  unsigned next_mcb_seg;
  unsigned prev_mcb_seg;
  unsigned paragraph_count;
  unsigned owner_psp;
  void __far *unused;
  // 0: normal memory block
  // 1: dta struct
  // 2: disk cache
  // 3: fs drv
  // 4: handle
  // 5: system
  unsigned char type; 
};


unsigned long get_first_mcb(void);
#pragma aux get_first_mcb = \
  "mov ah, 0x18" \
  "int 0x21" \
  value [dx ax];

unsigned free_memory(void);
#pragma aux free_memory = \
  "mov ah, 0x48" \
  "mov bx, -1" \
  "int 0x21" \
  value [bx] modify [ax];

#define GRAPH_NEWLINE 0x1000
#define SCREEN_COLS 80
#define GRAPH_DIV GRAPH_NEWLINE/(64)
unsigned memory_size(void);
#pragma aux memory_size = \
  "int 0x12" \
  value [ax]; 

unsigned this_psp(void);
#pragma aux this_psp = \
  "mov ax, cs" \
  value [ax];

inline unsigned empty_space_after(unsigned mcb)
{
  struct mcb __far *mcb_ptr = MK_FP(mcb, 0);
  unsigned mcb_end = mcb + 1 + mcb_ptr->paragraph_count;
  if(mcb_ptr->next_mcb_seg)
    return mcb_ptr->next_mcb_seg - mcb_end;
  return (memory_size()<<6) - mcb_end;
}

void bios_tt_putchar(char chr, unsigned char attr);
#pragma aux bios_tt_putchar = \
  "mov ah, 0x0E" \
  "xor bh, bh" \
  "int 0x10" \
  parm [al] [bl];

static void draw_cols(void)
{
  unsigned char n;
  unsigned char c;

  for(n=0,c=0;n<64;n++)
    if(!(n%4))
      printf(" %.4X", c++ << 12);
 

}

static void draw_mem(char chr)
{
  static unsigned full_chr = GRAPH_DIV-1;
  static unsigned par = 0;

  if(chr != '-')
    full_chr++;

  if(!(par % GRAPH_NEWLINE))
  {
    printf("%X", par>>12);
  }
  else if(!(par%(GRAPH_DIV*4)))
    putchar(' ');
  if(!(par%(GRAPH_DIV)))
  {
    if(full_chr)
      if(full_chr < ((GRAPH_DIV)>>2))
        chr = '/';
      else if(full_chr != GRAPH_DIV)
        if(chr != '-')
          chr += 32;
        else
          chr = '/';

    putchar(chr);
    full_chr = 0;
  }
  par++;
}

void show_memory_info(unsigned mcb, unsigned code_seg)
{
  unsigned caches = 0;
  unsigned system = 0;
  unsigned dtas = 0;
  unsigned programs = 0;
  unsigned fs = 0;
  unsigned kernel = mcb - code_seg;
  unsigned empty_space;
  unsigned char chr;
  unsigned n;
  struct mcb __far *mcb_ptr;
  unsigned my_psp = this_psp();
  unsigned max_paras = 0;
  unsigned cont_free_space = 0;
  struct file_handle __far *handle_ptr;

  mcb_ptr = MK_FP(mcb, 0);

  printf("Kernel: %u paragraphs (%u kilobytes), loaded to segment %.4Xh\n", kernel, kernel >> 6, code_seg);
  printf("Kernel heap: %u paragraphs (%u kilobytes)\n", mcb_ptr->paragraph_count, mcb_ptr->paragraph_count>>6);

  puts("Memory map: [B=BIOS K=Kernel P=Programs H=Handle D=DTA C=Cache F=FS\
 S=System]");
  puts("[/=Less than 1/4 used memory, Lowercase=Partially used memory]\n");

  draw_cols();

  for(n=0;n<code_seg;n++)
    draw_mem('B');
  for(n=0;n<kernel;n++)
    draw_mem('K');

  while(mcb)
  {
    mcb_ptr = MK_FP(mcb, 0);
    if(mcb != my_psp && mcb_ptr->owner_psp != my_psp)
    {
      if((mcb_ptr->type & 0x7F) != MCB_CACHE)
        cont_free_space = 0;
      else
        cont_free_space += mcb_ptr->paragraph_count+1;
      empty_space = 0;
      switch(mcb_ptr->type)
      {
        case MCB_NORMAL: default:
          programs += mcb_ptr->paragraph_count+1;
          chr = 'P';
        break;
        case MCB_DTA:
          dtas += mcb_ptr->paragraph_count+1;
          chr = 'D';
        break;
        case MCB_CACHE: case MCB_USERCACHE:
          caches += mcb_ptr->paragraph_count+1;
          chr = 'C';
        break;
        case MCB_FS:
          fs += mcb_ptr->paragraph_count+1;
          chr = 'F';
        break;
        case MCB_HANDLE:
          handle_ptr = MK_FP(mcb+1, 0);
          if(handle_ptr->owner_psp == my_psp)
            goto ownpsp2;
          programs += mcb_ptr->paragraph_count+1;
          chr = 'H';
        break;
        case MCB_SYSTEM:
          system += mcb_ptr->paragraph_count+1;
          chr = 'S';
      }
    }
    else
    {
      ownpsp2:
      empty_space = mcb_ptr->paragraph_count+1;
//      cont_free_space += empty_space;
      goto ownpsp;
    }

    for(n=0;n<mcb_ptr->paragraph_count+1;n++)
      draw_mem(chr);
    ownpsp:
    empty_space += empty_space_after(mcb);
    cont_free_space += empty_space;
    if(cont_free_space > max_paras)
      max_paras = cont_free_space;
    for(n=0;n<empty_space;n++)
      draw_mem('-');
    mcb = mcb_ptr->next_mcb_seg;
  }

  printf("\n\nSystem: %u paragraphs (%u kB), Caches: %u paragraphs (%u kB)\n", system, system>>6, caches, caches>>6);
  printf("Disk Transfer Areas: %u paragraphs (%u kB)\n", dtas, dtas>>6);
  printf("Programs: %u paragraphs (%u kB)\n", programs, programs>>6);
  printf("Filesystem driver: %u paragraphs (%u kB)\n", fs, fs>>6);
//  empty_space = free_memory();
  empty_space = max_paras;
  printf("Largest available free memory block: %u paragraphs (%u kB)\n", empty_space, empty_space>>6);
  printf("%u kilobytes total memory. ST-DOS MEM 1.11\n", memory_size());
  getchar();
}

int main()
{
  unsigned long mcb_info = get_first_mcb();

  show_memory_info((unsigned short)mcb_info, mcb_info>>16);

  return 0;
}
