#include <stdio.h>
#include <i86.h>
#include <dos.h>

unsigned char __based(__segname("_CODE")) error_code = 0;

#define PARALLEL_PORT_COUNT ((*(unsigned char __far*)0x00400011L) >> 6);

unsigned char __based(__segname("_CODE")) parallel_port_count;
unsigned char __based(__segname("_CODE")) highest_parallelport_number;

unsigned short __far * __based(__segname("_CODE")) parallelbase  = (unsigned short __far*)0x00400008L;

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];

struct pport_cb
{
  unsigned char row_len;
  unsigned char col_ptr;
  volatile unsigned char error;
};

static int is_invalid_parport(unsigned __far *port)
{
  if(!*port || (*port > highest_parallelport_number) || !parallelbase[--*port])
    return 1;
  return 0;
}

#define ERR_WAITFORACK 0x01
#define ERR_OUTOFPAPER 0x02
#define ERR_FATAL 0x03

struct pport_cb __far * __based(__segname("_CODE")) pport_cb;

int install_driver(char __far *, void __far *);
#pragma aux install_driver = \
  "mov ah, 0x6A" \
  "xor cx, cx" \
  "push ds" \
  "push es" \
  "pop ds" \
  "int 0x21" \
  "pop ds" \
  "jnc noerror" \
  "mov ax, -1" \
  "jmp end" \
  "noerror:" \
  "xor ax, ax" \
  "end:" \
  parm [dx bx] [es di] value [ax] modify [cx];


unsigned current_psp(void);
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

void __far *dos_alloc(unsigned bytes);
#pragma aux dos_alloc = \
  "push cx" \
  "mov cl, 4" \
  "add bx, 0xF" \
  "shr bx, cl" \
  "pop cx" \
  "mov ah, 0x48" \
  "int 0x21" \
  "jnc end" \
  "xor ax, ax" \
  "end:" \
  "xor dx, dx" \
  "xchg ax, dx" \
  parm [bx] value [dx ax];

void __far dos_free(unsigned seg);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

void debug(unsigned char chr);
#pragma aux debug = \
  "mov ah, 0x0E" \
  "mov bx, 0x0007" \
  "int 0x10" \
  parm [al] modify [bx];

void set_memory_allocation_strategy(unsigned strategy);
#pragma aux set_memory_allocation_strategy = \
  "mov ax, 0x5801" \
  "int 0x21" \
  parm [bx];

unsigned __far get_error(unsigned devnum);
#pragma aux get_error parm [bx] value [ax];
unsigned __far get_error(unsigned devnum)
{
  unsigned retval = error_code;
  error_code = 0;
  return retval;
}

unsigned __far lpt_write(unsigned, unsigned char __far*, unsigned);
#pragma aux lpt_write parm [bx] [dx ax] [cx] value [ax];

void wait_printer(unsigned long microseconds);
#pragma aux wait_printer = \
  "mov ah, 0x86" \
  "int 0x15" \
  "jnc end" \
  "inc cx" \
  "l1:" \
  "dec dx" \
  "jnz l1" \
  "loop l1" \
  "end:" \
  modify [ax] parm [cx dx];

static unsigned lpt_sendbyte(unsigned devnum, unsigned char byte)
{
  unsigned control_port = parallelbase[devnum]+2;
  unsigned printer_port = parallelbase[devnum];
  unsigned status_port = parallelbase[devnum]+1;
  unsigned retries = 0;

  if(inbyte(status_port) & 0x20)
  {
    error_code = 0x1C; // out of paper
    pport_cb[devnum].error |= ERR_OUTOFPAPER;
    return 1;
  }

  // wait while printer is busy
  while(++retries && !(inbyte(status_port) & 0x80))
    wait_printer(1);
  if(!retries) goto fatal_error;

  outbyte(printer_port, byte);

  // strobe on
  outbyte(control_port, 0x0D);

  wait_printer(10);

  // strobe off
  outbyte(control_port, 0x0C);

  while(!(inbyte(status_port) & 0x80) && retries++)
    wait_printer(1);

  if(!(inbyte(status_port) & 0x08))
  {
    fatal_error:
    error_code = 0x1F; // general failure
    pport_cb[devnum].error |= ERR_FATAL;
    return 2;
  }

  return 0;
}

unsigned __far lpt_write(unsigned devnum, unsigned char __far *buff, unsigned count)
{
  unsigned retval = 0;

  devnum--;

  while(count--)
  {
    if(lpt_sendbyte(devnum, *buff++))
      break;
    retval++;
    if(pport_cb[devnum].row_len)
    {
      if(++pport_cb[devnum].col_ptr == pport_cb[devnum].row_len)
      {
        if(lpt_sendbyte(devnum, '\r'))
          break;
        if(lpt_sendbyte(devnum, '\n'))
          break;
        pport_cb[devnum].col_ptr = 0;
      }
    }
  }

  return retval;
}

unsigned __far lpt_open(unsigned, unsigned);
#pragma aux lpt_open parm [bx] [cx] value [ax];
unsigned __far lpt_open(unsigned devnum, unsigned mode)
{
  unsigned n;
  if(is_invalid_parport(&devnum))
  {
    error_code = 1;
    return 1;
  }

  wait_printer(2000000);
  return 0;
}

void set_parallelport_settings(unsigned devnum, unsigned row_len)
{
  unsigned short control_port = parallelbase[devnum]+2;

  pport_cb[devnum].error = 0;
  pport_cb[devnum].col_ptr = 0;

  // disable interrupts
  outbyte(control_port, inbyte(control_port) & ~0x10);

  // reset the device and select the device and set direction to write
  outbyte(control_port, inbyte(control_port) & ~(0x04 | 0x20));
  wait_printer(1000);
  outbyte(control_port, inbyte(control_port) | 0x04 | 0x08);

  // if row_len != 0, switch auto line feed OFF
  if(pport_cb[devnum].row_len = row_len)
    outbyte(control_port, inbyte(control_port) & ~0x02);
  // else switch auto line feed ON
  else
    outbyte(control_port, inbyte(control_port) | 0x02);
}


unsigned __far driver_command(unsigned, unsigned char __far*);
#pragma aux driver_command parm [bx] [dx ax] value [ax];
unsigned __far driver_command(unsigned devnum, unsigned char __far *cmd)
{
  #define CMD_OPCODE cmd[0]
  #define CMD_ROWLEN cmd[1]

  if(is_invalid_parport(&devnum))
  {
    error_code = 0x1;
    return 1;
  }

  if(CMD_OPCODE)
    set_parallelport_settings(devnum, CMD_ROWLEN);

  CMD_ROWLEN = pport_cb[devnum].row_len;

  return 0;
}

unsigned long __far driver_ioctl(unsigned, unsigned, unsigned char __far*, unsigned);
#pragma aux driver_ioctl parm [bx] [cx] [dx ax] [si] value [dx ax];

#define CHAR_DEVICE 0x80L
#define NO_EOF_ON_INPUT 0x40L

unsigned long __far driver_ioctl(unsigned devnum, unsigned count,
                                 unsigned char __far *buff, unsigned function)
{
  unsigned status_port = parallelbase[devnum-1]+1;
  unsigned char status;

  switch(function & 0xFF)
  {
    case 0: case 1:
      return ((CHAR_DEVICE) << 16) | function;
    case 7:
      status = inbyte(status_port);
      // if out of paper, not ready or error, al = 0xFF
      if((status & 0x20) || !(status & (0x80 | 0x08)))
        return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
    case 6:
      // this device does not support reading, so al is always 0
      return ((long)FP_SEG(buff) << 16) | (function & 0xFF00);
  }
}


void __far * __based(__segname("_CODE")) function_table[] =
{
  0, // read
  lpt_write,
  0, // seek
  lpt_open, // open
  0, // close
  get_error, // get_error_code
  driver_command, // send_command
  driver_ioctl // IOCTL
};

int parallel_initialize(void)
{
  unsigned n, i;

  parallel_port_count = PARALLEL_PORT_COUNT;

  printf("%u parallel ports detected.\n", parallel_port_count);

  for(highest_parallelport_number=0,n=0,i=parallel_port_count;i;n++)
    if(parallelbase[n])
    {
      highest_parallelport_number = n+1;
      i--;
    }

  printf("Highest parallel port number is %u", highest_parallelport_number);

  if(highest_parallelport_number != parallel_port_count)
    printf(", but there are only %u parallel ports. Fix your jumper \
configuration.", parallel_port_count);


  pport_cb = dos_alloc(highest_parallelport_number*sizeof(struct pport_cb));
  if(!pport_cb) return -1;

  for(n=highest_parallelport_number;n--;)
    if(parallelbase[n])
    {
      pport_cb[n].col_ptr = 0;
      pport_cb[n].row_len = 80;

      set_parallelport_settings(n, 0);
    }

  puts("\nParallel ports initialized.");

  return 0;
}

int main()
{
  unsigned env_seg;
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  int rc;
  unsigned n;

  puts("Installing LPT driver.");

  set_memory_allocation_strategy(2);
  rc = parallel_initialize();
  set_memory_allocation_strategy(0);

  switch(rc)
  {
    case -2: puts("No parallel ports detected!"); return 1;
    case -1: puts("Out of memory!"); return 1;
  }

  if(!install_driver("LPT", function_table))
  {
    puts("Done, going resident...");

    env_seg = *(unsigned __far*)&psp_ptr[0x2C];
    dos_free(env_seg);

    for(n=3;n--;)
      dos_close(n);

    _dos_keep(0, 0x10 + ((FP_OFF(parallel_initialize) + 0xF) >> 4));
  }
  puts("Error");
  return 1;
}
